/*******************************************************************************

   eseries_kernel_interface.h
   Copyright 2016-2019 NetApp, Inc. All Rights Reserved.

   Provides definitions needed to access the eseries kernel interface module
   by application.

*******************************************************************************/

#ifndef __INCeseries_interface
#define __INCeseries_interface


/***  INCLUDES  ***/
#include "linux/ioctl.h"
#include "linux/types.h"

/***  CONSTANT DEFINITIONS  ***/
#define NUM_TSOD_DEVS          2

#define PR_SIGNATURE_LEN        32    /* persistent region signature length */
#define ERROR_REGION_SIZE      (8 * 1024)  /* arbitrarily chosen */
#define ERROR_LOGGING_SIZE     (ERROR_REGION_SIZE - 36)  /* Subtract signature and version */
#define NUM_ERR_ENTRIES        ( (ERROR_LOGGING_SIZE / sizeof(err_entry)) - 1 )

/***  TYPE DEFINITIONS  ***/

/* Data to be saved when disabling TSOD so that we restore to the prior state */
typedef struct
{
    uint32_t control[NUM_TSOD_DEVS];
    int      disabled[NUM_TSOD_DEVS];
} ESC_TSOD_STATE;

struct eki_eventfd_data
{
    int numVKICores;
    int* efd;
};

struct eki_eccinject_data
{
    uint64_t    physAddr;
    uint8_t     bit;
    uint8_t     immediate;
};

struct eki_ecc_check
{
    uint8_t     log_enabled;
    uint8_t     err_detected;
};

struct eki_addr_to_dimm
{
    uint32_t    dimm;
    uint64_t    addr;
};

struct eki_log_data
{
    uint64_t buffer;       /* IN - guest addr of buffer */
    uint8_t  all_entries;  /* IN - 0=active entries, 1=all entries */
    uint32_t nr_entries;   /* OUT - num entries returned */
};

struct eki_log_ack
{
    uint32_t id;        /* IN - event id to ack */
    uint32_t nr_acked;  /* OUT - number of events ack'd */
};

struct eki_mmconfig
{
    uint64_t    mmconfigAddr;
};

struct mce_error {
    uint64_t maddr;
    uint64_t status;
    uint64_t misc;
    uint16_t corr_err_cnt;
    uint8_t  msr_bank;
    uint8_t  dimm;
    uint8_t  is_sb_enabled;
    uint8_t  is_mb_enabled;
    uint8_t  is_uncorrectable;
    uint8_t  is_eos;
    uint8_t  is_rpa;
    uint8_t  is_cache_msr;
}; /* 33 bytes */

/* ECC errors can be routed through the Home Agent (HA) or Mem Channel (MC) MSR banks.
 * Some errors (MB triggered via CPU) will only have info in the HA regs.
 * Some errors (any triggered via patrol scrub) will only have info in the MC regs.
 * Some errors (SB triggered via CPU) will have top-level info in the HA regs and
 * secondary info in the MC regs.
 * Errors that have info in both the HA and MC regs will put the HA info in the
 * maddr/status/misc/bank variables and the MC info in the sec_addr/status/misc/bank vars.
 * All other errors will just put the relevant HA/MC info in the maddr/status/misc vars.
 */
struct ecc_error {
    uint64_t maddr;
    uint64_t status;
    uint64_t misc;
    uint64_t sec_addr;
    uint64_t sec_status;
    uint64_t sec_misc;
    uint32_t corr_bitmask; /* Only valid for correctable errors */
    uint16_t corr_err_cnt; /* Only valid for correctable errors */
    uint8_t  bank;         /* Bank # that maddr/status/misc came from */
    uint8_t  sec_bank;     /* Bank # that sec_addr/status/misc came from */
    uint8_t  dimm;
    uint8_t  is_sb_enabled;
    uint8_t  is_mb_enabled;
    uint8_t  is_uncorrectable;
    uint8_t  is_eos;
    uint8_t  is_rpa;
    uint8_t  err_device;
}; /* 63 bytes */

/*
 * pci types
 * PCI errors are currently not used, but this was
 * ported from Xen in case it is used in the future.
 * copied from xen/include/xen/pci.h
 */
#define PCI_ERR_UNKNOWN     0
#define PCIe_ENDPOINT       1
#define PCIe_BRIDGE         2 // PCIe root port, switch
#define PCIe2PCI_BRIDGE     3 // PCIe-to-PCI/PCIx bridge
#define PCI2PCIe_BRIDGE     4 // PCI/PCIx-to-PCIe bridge
#define LEGACY_PCI_BRIDGE   5 // Legacy PCI bridge
#define PCI_HOST_BRIDGE     6 // PCI Host bridge
#define PCI                 7

struct pci_error {
    uint8_t  pci_type;
    uint8_t  bus;
    uint8_t  device;
    uint8_t  func;
    uint8_t  is_uncorrectable;
    uint8_t  needs_lockdown;
    uint16_t vid;
    uint16_t did;
    uint16_t pci_status;
    uint16_t pci_sec_status;
    uint16_t pcie_type;
    uint16_t pcie_dev_status;
    uint16_t pcie_slot_status;
    uint16_t pcie_link_status;
    uint16_t pcie_link_status2;
    uint32_t pcie_link_cap;
    uint32_t aer_cor_status;
    uint32_t aer_unc_status;
    uint32_t aer_root_status;
    uint32_t xp_cor_status;
    uint32_t xp_unc_status;
    uint32_t xp_gbl_status;
    uint32_t header_log[4];
}; /* 70 bytes */

/*
 * Status registers that may be useful for a given
 * error event but not necessary to log in each entry
 * (one entry per event instead of multiple copies
 * for multiple entries)
 */
struct reg_error {
    uint32_t gn_status;         /* GNERRST */
    uint32_t gf_status;         /* GFERRST */
    uint32_t gc_status;         /* GCERRST */
    uint32_t irpp0_status;      /* IRPP0ERRST */
    uint32_t irpp1_status;      /* IRPP1ERRST */
    uint32_t iio_status;        /* IIOERRST */
    uint32_t mi_status;         /* MIERRST */
    uint32_t vtd_status;        /* VTUNCERRSTS */
    uint8_t  is_uncorrectable;
}; /* 33 bytes */

/* error entry states */
#define ERROR_ENTRY_EMPTY   0
#define ERROR_ENTRY_WRITING 1
#define ERROR_ENTRY_USED    2

/* error_entry types */
#define ERROR_TYPE_MCE      1
#define ERROR_TYPE_PCI      2
#define ERROR_TYPE_REG      3
#define ERROR_TYPE_ECC      4

/*
 * Current error entry data type version
 * This value is stored in the error logging when it is initialized.
 * Increment this version when the error log structures are changed
 * to force re-initialization.
 */
#define ERROR_ENTRY_SIZE        128     /* for compile-time checking */
#define ERROR_ENTRY_DATA_SIZE   (ERROR_ENTRY_SIZE - 16)
typedef struct error_entry {
    uint32_t id;
    uint8_t  state;
    uint8_t  type;
    uint16_t pad1;
    uint64_t timestamp;
    union {
        struct mce_error mce;
        struct pci_error pci;
        struct reg_error reg;
        struct ecc_error ecc;

        /* maximum size of any of above union members */
        struct { uint8_t pad[ERROR_ENTRY_DATA_SIZE]; };
    };
} err_entry;

/*
 * Using this structure persistent region imt tests can read/write the
 * version and signature of the header region and error logging region.
 */
struct persistent_reg_imtdata {
   enum {
      READ_GLOBAL,
      READ_REGION,
      WRITE_GLOBAL,
      CALL_INIT
    } op;                                    /* what operation to perform */

   enum {
      REG_VER,
      REG_SIG,
      ERR_REG_VER,
      ERR_REG_SIG,
    } var;                                   /* on which variable */

   union {
      uint32_t version;
      char     signature[PR_SIGNATURE_LEN];
    } data;                                 /* to store the data read or to be written */

};/* 40 bytes */


/***  MACRO DEFINITIONS  ***/
/*
 * IOCTL codes
 * - magic number of 'e' is selected for EKI IOCTL
 * - ioctl index numbers are starting with 1
 */
#define EKI_IOCTL_MAGIC_NUMBER 'e'

/* 1-10 are reserved for NMI */
#define EKI_IOCTL_NMI_SEND_FDS       _IOW(EKI_IOCTL_MAGIC_NUMBER,   1, struct eki_eventfd_data)

/* 11-20 are reserved for ECC */
#define EKI_IOCTL_ECC_INJECT         _IOW(EKI_IOCTL_MAGIC_NUMBER,  11, struct eki_eccinject_data)
#define EKI_IOCTL_ECC_SCRUB_ENABLE   _IOW(EKI_IOCTL_MAGIC_NUMBER,  12, int)
#define EKI_IOCTL_ECC_ENABLE_MB      _IOW(EKI_IOCTL_MAGIC_NUMBER,  13, int)
#define EKI_IOCTL_ECC_CHECK          _IOWR(EKI_IOCTL_MAGIC_NUMBER, 14, struct eki_ecc_check)
#define EKI_IOCTL_ADDR_TO_DIMM       _IOWR(EKI_IOCTL_MAGIC_NUMBER, 15, struct eki_addr_to_dimm)

/* 21-30 are reserved for logging */
#define EKI_IOCTL_LOG_GET            _IOWR(EKI_IOCTL_MAGIC_NUMBER, 21, struct eki_log_data)
#define EKI_IOCTL_LOG_ACK            _IOWR(EKI_IOCTL_MAGIC_NUMBER, 22, struct eki_log_ack)
#define EKI_IOCTL_LOG_DUMP           _IO(EKI_IOCTL_MAGIC_NUMBER,   23)
#define EKI_IOCTL_LOG_EVENT          _IOW(EKI_IOCTL_MAGIC_NUMBER,  24, uint64_t)

/* 31-40 are reserved for miscellaneous */
#define EKI_IOCTL_MMCONFIG_GET       _IOR(EKI_IOCTL_MAGIC_NUMBER,  31, struct eki_mmconfig)
#define EKI_IOCTL_CPUFREQ_GET        _IOR(EKI_IOCTL_MAGIC_NUMBER,  32, uint64_t)
#define EKI_IOCTL_REL_CPUTEMP_GET    _IOR(EKI_IOCTL_MAGIC_NUMBER,  33, uint8_t)
#define EKI_IOCTL_TCCTEMP_GET        _IOR(EKI_IOCTL_MAGIC_NUMBER,  34, uint8_t)
#define EKI_IOCTL_PCI_SHUTDOWN       _IO(EKI_IOCTL_MAGIC_NUMBER,   35)
#define EKI_IOCTL_TSOD_DISABLE       _IOWR(EKI_IOCTL_MAGIC_NUMBER, 36, ESC_TSOD_STATE)
#define EKI_IOCTL_TSOD_RESTORE       _IOW(EKI_IOCTL_MAGIC_NUMBER,  37, ESC_TSOD_STATE)
#define EKI_IOCTL_WB_INVALIDATE      _IO(EKI_IOCTL_MAGIC_NUMBER,   38)

/* 41-50 are reserved for IMTs */
#define EKI_IOCTL_PER_REG_IMT        _IOWR(EKI_IOCTL_MAGIC_NUMBER, 41, struct persistent_reg_imtdata)


/* eseries kernel interface driver name */
#define EKI_DEV_NAME            "eki"
/* eseries kernel interface driver path */
#define EKI_DEV_PATH            "/dev/" EKI_DEV_NAME

/***  GLOBALS  ***/

#endif        /* End of __INCeseries_interface */

