/*******************************************************************************

   ecc.c
   Copyright 2019 NetApp, Inc. All Rights Reserved.

   ECC handling general infrastructure

*******************************************************************************/
#include <linux/types.h>
#include <linux/notifier.h>
#include <asm/mce.h>
#include <asm/msr.h>

#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,16,0)
    #include <linux/sched/signal.h>
#else
    #include <linux/sched.h>
#endif

#include "ecc.h"
#include "eccP.h"
#include "eki_common.h"
#include "rmem-mgr.h"

bool mce_in_progress = false;

ECC_INTERFACE* ecc_iface = 0;

static unsigned long eosAddr = 0;
static unsigned long eosLength = 0;
static unsigned long rpaAddr = 0;

/* Number of MCE banks */
static int mceBanksCnt = 0;


/**
 * \brief Get the eos and rpa memory ranges from rmem.  The eos range should
 *        always be set, but the rpa ranges are not accessible until the EOS
 *        application is running.
 *
 */
static void
get_eos_addr_ranges
    (
    void
    )
{
    /* Obtain eos address ranges from rmem unless they have been retrieved */
    if (!eosAddr || !eosLength || !rpaAddr)
    {
        rmem_eos_rpa_memory(&eosAddr, &eosLength, &rpaAddr);
        mce_printk(KERN_WARNING, "EOS Addr 0x%lx, EOS Length 0x%lx, RPA Addr 0x%lx\n",
                                 eosAddr, eosLength, rpaAddr);
    }
}

/**
 * \brief  Determine if the address is within EOS memory.
 *
 * \param[in]  addr   Address to check to see if in the range of EOS memory
 *
 * \return 1 if it is in EOS memory and 0 otherwise.
 */
int
is_eos_mem
    (
    uint64_t addr
    )
{
    get_eos_addr_ranges();
    /* Note: EOS memory includes RPA memory */
    if ( ((unsigned long)addr >= eosAddr) &&
         ((unsigned long)addr < (eosAddr + eosLength)) )
        return 1;
    return 0;
}

/**
 * \brief  Determine if the address is within RPA memory.
 *
 * \param[in]  addr   Address to check to see if in the range of RPA memory
 *
 * \return 1 if it is in RPA memory and 0 otherwise.
 */
int
is_rpa_mem
    (
    uint64_t addr
    )
{
    get_eos_addr_ranges();
    /* Note: RPA memory continues to the end of EOS memory */
    if ((addr >= rpaAddr) && (addr < (eosAddr + eosLength)))
        return 1;
    return 0;
}

/**
 * \brief  Command to enable/disable patrol scrub
 *
 * \param[in]  enable   Boolean indicate whether to enable or disable patrol scrub
 *
 * \return 1 patrol scrub was previously enabled and 0 otherwise
 */
bool
enable_patrol_scrub
    (
    bool enable
    )
{
    if (!ecc_iface)
    {
        printk(KERN_ERR "%s: ecc_iface is not initialize\n", __func__);
        return 0;
    }
    return ecc_iface->enable_patrol_scrub(enable);
}

/**
 * \brief  Control the cache by setting/clearing the CD (cache disable) bit in
 *         the CR0 register.
 *
 * \param[in]  enable   Boolean indicate whether to enable or disable cache
 */
void
cache_enable
    (
    bool enable
    )
{
    unsigned long cr0;

    cr0 = read_cr0();

    if (enable)
    {
        cr0 &= ~X86_CR0_CD;
        write_cr0(cr0);
    }
    else
    {
        cr0 |= X86_CR0_CD;
        write_cr0(cr0);
        wbinvd();
    }
}

/**
 * \brief   ECC error injection routine. Injects a single or multi bit error at a given
 *          machine address. It temporarily disables the HW patrol scrub so no latent
 *          errors are detected during injection (and so we don't get unlucky and inject
 *          an error where the patrol scan is checking). Caching is disabled to force a
 *          write to go out to physical memory.
 *
 * \param[in]  maddr      Address where ECC error should be injected
 * \param[in]  bit        Which bit should get corrupted (ECC_MULTI_BIT for multibit errors)
 * \param[in]  immediate  Flag when set to trigger the error right after injecting condition
 */
void
ecc_inject
    (
    uint64_t maddr,
    uint8_t bit,
    uint8_t immediate
    )
{
    if (!ecc_iface)
    {
        printk(KERN_ERR "%s: ecc_iface is not initialize\n", __func__);
        return;
    }
    ecc_iface->ecc_inject(maddr, bit, immediate);
}

/**
 * \brief   Common print routine to ensure consistency when printing/logging
 *
 * \param[in]  bank       Which bank error occurred on
 * \param[in]  status     Status from the error
 * \param[in]  addr       Address where error happened
 * \param[in]  misc       Misc data from error
 */
void
print_mce_msg
    (
    int bank,
    uint64_t status,
    uint64_t addr,
    uint64_t misc
    )
{
    mce_printk(KERN_WARNING, "Machine check, Bank %2d: Status is 0x%llx\n", bank, status );
    mce_printk(KERN_WARNING, "                       Address is 0x%llx\n", addr );
    mce_printk(KERN_WARNING, "                          Misc is 0x%llx\n", misc );
}

/**
 * \brief   Common print routine to close out error to ensure consistency when printing/logging
 *
 * \param[in]  status     Status from the error
 */
void
print_mce_end_msg
    (
    uint32_t status
    )
{
    if (status & (ECC_UNCOR | MCE_UNCOR))
    {
        mce_printk(KERN_WARNING, "UNCORRECTABLE ERROR!\n");
    }
    else if (status & (ECC_COR | MCE_COR))
    {
        mce_printk(KERN_WARNING, "CORRECTABLE ERROR!\n");
    }
    else if (status & MCE_UNKNOWN)
    {
        /* We should never have an unknown status - if we do it is some strange
         * hardware state that we did not expect. */
        mce_printk(KERN_ERR, "UNKNOWN ERROR!!!\n");
    }
    else
    {
        /* Should never reach here - if we do, we have a bug in our code */
        mce_printk(KERN_ERR, "%s: No error detected\n", __func__);
    }
}

/**
 * \brief Machine Check Handler
 *
 * This routine is called by the kernel when a machine check occurs.
 *
 */
static void
machine_check_handler
    (
    void
    )
{
    uint64_t gstatus;
    uint32_t check_sts = 0;

    mce_in_progress = true;

    /* get MCIP, EIPV & RIPV bits */
    rdmsrl(MSR_IA32_MCG_STATUS, gstatus); /* get MCIP, EIPV & RIPV bits */

    check_sts = check_for_errors(true);

    /* Fatal error, panic */
    if (check_sts & (ECC_UNCOR | MCE_UNCOR | MCE_UNKNOWN))
    {
        panic("EKI: Fatal machine check on current CPU");
    }

    /* Invalid IP, don't continue */
    if (!(gstatus & MCG_STATUS_RIPV))
    {
        force_sig(SIGBUS, current);
    }

    /* Non-fatal error, continue */
    mce_in_progress = false;
}

/**
 * \brief  Command to clear errors for specified bank
 *
 * \param[in]  bank   Bank in which to clear the errors
 */
void
clear_bank_errors
    (
    int banknum
    )
{
    uint64_t status;

    rdmsrl(MSR_IA32_MCx_STATUS(banknum), status);

    if (status & MCI_STATUS_ADDRV)
        wrmsrl(MSR_IA32_MCx_ADDR(banknum), 0x0ULL);
    if (status & MCI_STATUS_MISCV)
        wrmsrl(MSR_IA32_MCx_MISC(banknum), 0x0ULL);

    wrmsrl(MSR_IA32_MCx_STATUS(banknum), 0x0ULL);
}

/**
 * \brief   Function to report number of MCE channels
 *
 * \return  Number of MSR banks related to MCE handling
 */
int
get_mce_bank_cnt
    (
    void
    )
{
    return mceBanksCnt;
}

/**
 * \brief   Function initial MCE handling
 *
 */
static void
mce_init
    (
    void
    )
{
    uint64_t msr_content;

    rdmsrl(MSR_IA32_MCG_CAP, msr_content);

    if (msr_content & MCG_CTL_P) /* Control register present ? */
        wrmsrl(MSR_IA32_MCG_CTL, 0xffffffffffffffffULL);

    mceBanksCnt = msr_content & MCG_BANKCNT_MASK;
    printk(KERN_INFO "Number of mce banks: %d\n", mceBanksCnt);
}

/**
 * \brief   Wrapper function to call processor specific routine to check for errors
 *
 * \param[in]  enable_log     Enable writing to error log
 *
 * \return     Non-zero value if errors were detected
 */
uint32_t
check_for_errors
    (
    bool enable_log
    )
{
    if (!ecc_iface)
    {
        printk(KERN_ERR "%s: ecc_iface is not initialize\n", __func__);
        return 0;
    }
    return ecc_iface->check_for_errors(enable_log);
}

/**
 * \brief   Wrapper function to call processor specific routine to enable/disable
 *          MCE for multibit errors.
 *
 * \param[in]  enable     Enable/Disable multibit error reporting
 */
void
mce_enable_mb
    (
    bool enable
    )
{
    if (!ecc_iface)
    {
        printk(KERN_ERR "%s: ecc_iface is not initialize\n", __func__);
        return;
    }
    ecc_iface->mce_enable_mb(enable);
}

/**
 * \brief   Wrapper function to call processor specific routine to check whether
 *          MCE for multibit errors is enabled/disabled.
 *
 * \return  1=Enabled  0=Disabled
 */
bool
mce_mb_enabled
    (
    void
    )
{
    if (!ecc_iface)
    {
        printk(KERN_ERR "%s: ecc_iface is not initialize\n", __func__);
        return 0;
    }
    return ecc_iface->mce_mb_enabled();
}

/**
 * \brief   Wrapper function to call processor specific routine to clear memory block
 *
 * \param[in]  addrStart     Starting address for clearing memory
 * \param[in]  length        Number of bytes to clear
 * \param[in]  logErrors     Flag indicating whether to log errors during clearing
 */
void
ecc_clear_memory
    (
    void* addrStart,
    unsigned long length,
    bool logErrors
    )
{
    uint32_t errStatus;
    bool patrol_status;
    bool mb_status;

    if (!ecc_iface)
    {
        printk(KERN_ERR "%s: ecc_iface is not initialize\n", __func__);
        return;
    }

    if (!addrStart)
    {
        printk(KERN_INFO "Clearing memory: NULL address\n");
        return;
    }

    mb_status = ecc_iface->mce_mb_enabled();

    /* Disable ECC error detection */
    patrol_status = ecc_iface->enable_patrol_scrub(false);
    ecc_iface->mce_enable_mb(false);

    /* Clear memory */
    memset(addrStart, 0, length);

    /* Flush cache */
    wbinvd();

    /* Clear any ECC errors */
    errStatus = ecc_iface->check_for_errors(logErrors);
    printk(KERN_INFO "Clearing memory: %p/%lu - errors=%u\n",
                      addrStart, length,  errStatus);

    /* Restore ECC error detection */
    ecc_iface->mce_enable_mb(mb_status);
    ecc_iface->enable_patrol_scrub(patrol_status);
}

/**
 * \brief   Wrapper function to call processor specific routine to convert
 *          address to dimm number
 *
 * \param[in]  addr     Address to convert to dimm
 *
 * \return  DIMM index
 */
int32_t
maddr_to_dimm
    (
    uint64_t addr
    )
{
    if (!ecc_iface)
    {
        printk(KERN_ERR "%s: ecc_iface is not initialize\n", __func__);
        return -1;
    }
    return ecc_iface->maddr_to_dimm(addr);
}


/*************************************************************************
 *   Default ECC error handling
 *************************************************************************
 */


/**
 * \brief   Stub routine for enable_patrol_scrub
 *
 * \param[in]  enable   Boolean indicate whether to enable or disable patrol scrub
 *
 * \return 1 patrol scrub was previously enabled and 0 otherwise
 */
static bool
enable_patrol_scrub_default
    (
    bool enable
    )
{
    (void)enable;
    return 0;
}

/**
 * \brief   Stub routine for ECC error injection.
 *
 * \param[in]  maddr      Address where ECC error should be injected (ignored)
 * \param[in]  bit        Which bit should get corrupted (ECC_MULTI_BIT for multibit errors) (ignored)
 * \param[in]  immediate  Flag when set to trigger the error right after injecting condition (ignored)
 */
static void
ecc_inject_default
    (
    uint64_t maddr,
    uint8_t bit,
    uint8_t immediate
    )
{
    (void)maddr;
    (void)bit;
    (void)immediate;
    mce_printk(KERN_INFO, "ECC injection is not supported for this processor\n");
}

/**
 * \brief   Stub routine for check_for_errors.
 *
 * \param[in]  enable_log     Enable writing to error log (ignored)
 *
 * \return     Alway returns zero
 */
static uint32_t
check_for_errors_default
    (
    bool enable_log
    )
{
    static bool stopPrinting = false;
    int bankIdx;
    uint64_t status;
    uint32_t returnSts = 0;

    (void)enable_log;

    if (!stopPrinting)
    {
        mce_printk(KERN_ERR, "Minimal support to check for ECC/MCE for this processor\n");
        stopPrinting = true;
    }

    /* For unsupported platforms, we simply clear the errors
     * and report back any uncorrectable errors
     */
    for (bankIdx = 0; bankIdx < get_mce_bank_cnt(); ++bankIdx)
    {
        rdmsrl(MSR_IA32_MCx_STATUS(bankIdx), status);

        /* No data to report */
        if (!(status & MCI_STATUS_VAL))
        {
            continue;
        }

        if (status & MCI_STATUS_UC)
        {
            returnSts |= MCE_UNCOR;
        }

        clear_bank_errors(bankIdx);
    }

    return returnSts;
}

/**
 * \brief   Stub routine for enabling MB errors
 *
 * \param[in]  enable     Enable/Disable multibit error reporting (ignored)
 */
static void
mce_enable_mb_default
    (
    bool enable
    )
{
    (void)enable;
}

/**
 * \brief   Stub routine for MB enabled state
 *
 * \return  0=Disabled
 */
static bool
mce_mb_enabled_default
    (
    void
    )
{
    return 0;
}

/**
 * \brief   Stub routine for maddr_to_dimm
 *
 * \param[in]  addr     Address to convert to dimm
 *
 * \return  DIMM index
 */
static int32_t
maddr_to_dimm_default
    (
    uint64_t addr
    )
{
    (void)addr;
    return -1;
}

/**
 * \brief   Stub routine for initializing driver
 *
 */
static void
init_default
    (
    void
    )
{
}

static ECC_INTERFACE ecc_iface_default =
{
    .enable_patrol_scrub = enable_patrol_scrub_default,
    .ecc_inject          = ecc_inject_default,
    .mce_mb_enabled      = mce_mb_enabled_default,
    .mce_enable_mb       = mce_enable_mb_default,
    .check_for_errors    = check_for_errors_default,
    .maddr_to_dimm       = maddr_to_dimm_default,
    .init                = init_default
};


/**
 * \brief   Top level routine for initializing ECC error handling.  This
 *          routine performs registration of ECC eror clearing then
 *          it determines the right driver for the processor and
 *          calls its initialization routine.
 *
 * \return     0 = successful,  non-zero = failure
 */
int
ecc_init
    (
    void
    )
{
    int ret;

    /* Provide a function to the rmem manager to clear EOS heap memory */
    rmem_register_clear_memory(ecc_clear_memory);

    switch (cpu)
    {
        case BROADWELL:
        {
            extern ECC_INTERFACE ecc_iface_broadwell;
            ecc_iface = &ecc_iface_broadwell;
            break;
        }
        case SKYLAKE:
        {
            extern ECC_INTERFACE ecc_iface_skylake;
            ecc_iface = &ecc_iface_skylake;
            break;
        }
        default:
            ecc_iface = &ecc_iface_default;
            break;
    }

    /* Initialize MCE handling */
    mce_init();

    ecc_iface->init();

    printk(KERN_INFO "ECC Init: setting machine check vector\n");
    ret = register_mce_handler(machine_check_handler);
    if (ret != 0)
        printk(KERN_ERR "ECC Init: machine check registration failed\n");

    return ret;
}
