/*******************************************************************************

   ecc_broadwell.c
   Copyright 2016-2019 NetApp, Inc. All Rights Reserved.

   ECC handling for the broadwell processor

*******************************************************************************/
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/mm.h>
#include <linux/mm_types.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/uaccess.h>
#include <asm/mce.h>
#include <asm/msr.h>
#include <asm/tlbflush.h>
#include <asm/paravirt.h>
#include <asm/page_types.h>

#include "ecc.h"
#include "eccP.h"
#include "logging.h"


/*
 * Various definitions used to decode a system address to a DIMM.
 */
#define NUM_TAD_REGS         12
#define ADDR_BIT_45_26_MSK   0x3FFFFC000000ULL
#define TAD_RULE_LIMIT_SHIFT 14

/* SCRUBCTL - SCRUB_EN bit */
#define MEM_CTRL_SCRUB_EN   0x80000000

/* SYSDEFEATURE2 - MCCHAN SHIFT UP ENABLE */
#define MCCHAN_SHIFT_UP_ENABLE (1 << 22)

/* SYSDEFEATURE2 - MCCHAN HASH ENABLE */
#define MCCHAN_HASH_ENABLE (1 << 21)

/* DIMM indeterminate */
#define DIMM_INDETERM   0xFF

/* MSR bank mappings for Home Agent (HA) and Mem Channels */
struct ecc_banks
{
    int home_agent;
    int mem_channel[4];
    int dimm[4];
};

/* Broadwell: Only has 2 channels: Banks 9, 10 (0's == No Bank) */
static struct ecc_banks broadwell_de_ecc_banks =
    { 7,
        { 9, 10, 0, 0 },
        {0, 1, DIMM_INDETERM, DIMM_INDETERM}
    };

/* Common print routine to ensure consistency when printing/logging */
/* Various registers that we access via uncore_reg_read/write */
enum uncore_reg {
    TAD_RULE,
    TADWAY,
    MEM_TECH,
    SCRUBCTL,
    ADDR_MATCH_LO,
    ADDR_MATCH_HI,
    ADDR_MASK_LO,
    ADDR_MASK_HI,
    DEV0_XOR_MASK,
    DEV1_XOR_MASK,
    INJ_EXTRA,
    SYSDEFEATURE2
};

static DEFINE_SPINLOCK(error_inj_lock);


/***   FORWARD DECLARATIONS ***/

static const struct ecc_banks* get_mce_ecc_banks(void);
static uint32_t check_for_errors_broadwell(bool enable_log);
static bool mce_mb_enabled_broadwell(void);


/* Get Memory Controller Configuration register B:D:F/addr for a given
 * register/processor.
 * See Broadwell-DE EDS Vol 2 for register details.
 * SCRUBCTL - Addr 0xA0
 *   Broadwell DE:      0xFF:19:1
 * ADDR_MATCH_LO - Addr 0x1C0
 * ADDR_MATCH_HI - Addr 0x1C4
 * ADDR_MASK_LO  - Addr 0x1C8
 * ADDR_MASK_HI  - Addr 0x1CC
 *   Broadwell DE:      (channel 0) 0xFF:19:2
 *   Broadwell DE:      (channel 1) 0xFF:19:3
 */
static void
mc_reg_info
    (
    enum uncore_reg reg_name,
    uint32_t channel,
    unsigned int *dev,
    unsigned int *func,
    unsigned int *addr
    )
{
    switch (reg_name)
    {
        case TADWAY:
        case MEM_TECH:
            *addr = 0x80;
            break;
        case SCRUBCTL:
            *addr = 0xA0;
            break;
        case ADDR_MATCH_LO:
            *addr = 0x1C0;
            break;
        case ADDR_MATCH_HI:
            *addr = 0x1C4;
            break;
        case ADDR_MASK_LO:
            *addr = 0x1C8;
            break;
        case ADDR_MASK_HI:
            *addr = 0x1CC;
            break;
        default:
            *addr = 0;
            *dev = 0;
            *func = 0;
            return;
    }

    *dev = 19;

    /*
     * TADWAY is in function 0.
     * SCRUBCTL is in function 1.
     * The rest of the MC regs are channel dependent (channel # + 2).
     */
    switch (reg_name)
    {
        case TADWAY:
            *func = 0;
            break;
        case SCRUBCTL:
            *func = 1;
            break;
        default:
            *func = channel + 2;
            break;
    }
}

/* Get Memory Controller Error Injection register B:D:F/addr for a given
 * register/processor.
 * See Broadwell-DE EDS Vol 2 for register details.
 * DEV0_XOR_MASK - Addr 0x200
 * DEV1_XOR_MASK - Addr 0x204
 * INJ_EXTRA     - Addr 0x208
 *   Broadwell DE:      (channel 0) 0xFF:20:2
 *   Broadwell DE:      (channel 1) 0xFF:20:3
 */
static void
inj_reg_info
    (
    enum uncore_reg reg_name,
    uint32_t channel,
    unsigned int *dev,
    unsigned int *func,
    unsigned int * addr
    )
{
    switch (reg_name)
    {
        case DEV0_XOR_MASK:
            *addr = 0x200;
            break;
        case DEV1_XOR_MASK:
            *addr = 0x204;
            break;
        case INJ_EXTRA:
            *addr = 0x208;
            break;
        default:
            *addr = 0;
            *dev = 0;
            *func = 0;
            return;
    }

    *dev = 20;

    if (channel == 0)
        *func = 2;
    else
        *func = 3;
}

/* Get Memory Controller Error Injection register B:D:F/addr for a given
 * register/processor.
 * See Broadwell-DE EDS Vol 2 for register details.
 * SYSDEFEATURE2 - Addr 0x84
 */
static void
ha_reg_info
    (
    enum uncore_reg reg_name,
    unsigned int *dev,
    unsigned int *func,
    unsigned int * addr
    )
{
    switch (reg_name)
    {
        case SYSDEFEATURE2:
            *dev = 18;
            *func = 0;
            *addr = 0x84;
            break;
        default:
            *addr = 0;
            *dev = 0;
            *func = 0;
            return;
    }
}

/* Generic access function for all uncore registers. Currently there are
 * three categories of registers: TAD registers, Memory Controller registers,
 * and Error Injection registers. All of these are accessed in PCI config
 * space. */
static uint32_t
uncore_reg_rw
    (
    enum uncore_reg reg_name,
    uint32_t index,
    uint32_t value,
    bool     write
    )
{
    unsigned int dev=0;
    unsigned int func=0;
    unsigned int addr=0;
    struct   pci_dev *pdev;

    switch (reg_name)
    {
        case TAD_RULE:
            /* Get the Target Address Decode DRAM Rule register info for a given
             * register/processor. There are 12 TAD registers that contain various
             * addressing/configuration info.
             * See Broadwell-DE EDS Vol 2 for register details.
             * Broadwell-DE: 0xFF:18:0 - Addr 0x40, 0x44, 0x48...0x6C
             */
            dev = 18;
            addr = 0x40 + 4*index;
            func = 0;
            break;
        case TADWAY:
        case MEM_TECH:
        case SCRUBCTL:
        case ADDR_MATCH_LO:
        case ADDR_MATCH_HI:
        case ADDR_MASK_LO:
        case ADDR_MASK_HI:
            mc_reg_info(reg_name, index, &dev, &func, &addr);
            break;
        case DEV0_XOR_MASK:
        case DEV1_XOR_MASK:
        case INJ_EXTRA:
            inj_reg_info(reg_name, index, &dev, &func, &addr);
            break;
        case SYSDEFEATURE2:
            ha_reg_info(reg_name, &dev, &func, &addr);
            break;
        default:
            /* Should never reach here, but need to keep compiler happy.
             * Just return a bogus value. */
            return 0xFFFFFFFF;
            break;
    }

    pdev = pci_get_domain_bus_and_slot(0, 0xFF, PCI_DEVFN(dev, func));
    if (pdev)
    {
        if (write)
        {
            pci_write_config_dword(pdev, addr, value);
            return 0;
        }
        else
        {
            pci_read_config_dword(pdev, addr, &value);
            return value;
        }
    }
    else
    {
        return 0xFFFFFFFF;
    }
}

/* Read wrapper for uncore_reg_rw */
static uint32_t
uncore_reg_read
    (
    enum uncore_reg reg_name,
    uint32_t index
    )
{
    return uncore_reg_rw(reg_name, index, 0, 0);
}

/* Write wrapper for uncore_reg_rw */
static void
uncore_reg_write
    (
    enum uncore_reg reg_name,
    uint32_t index,
    uint32_t value
    )
{
    (void)uncore_reg_rw(reg_name, index, value, 1);
}

/* Convert machine address to 0-indexed Channel #.
 * The TAD registers are used to determine which addrs
 * are mapped to specific channels. Check each TAD register to see if a given
 * address is covered by it. If so, then check to see how many channels are
 * in use. Then figure out which of those channels the original address belongs to.
 */
static int
maddr_to_channel
    (
    uint64_t addr
    )
{
    uint64_t tadLimit;
    uint32_t index;
    uint32_t tadReg;
    uint32_t chInterleave;
    uint32_t wayness;

    /*
     *  The following algorithm was extracted from the Intel cScripts
     *  file for the broadwell DE processor.
     *  The specific function referenced is in bdxAddressTranslator.py
     *  function forwardTad.
     */

    uint32_t sysDefeature2 = uncore_reg_read(SYSDEFEATURE2, 0);

    /*
     * Number of bits to shift address to get initial channel interleave index
     * It is either 6 or 7 depending on whether the shift enable is set
     */
    uint32_t shiftCnt = (sysDefeature2 & MCCHAN_SHIFT_UP_ENABLE)? 7: 6;

    /* Find the TAD rule corresponding to the given address. */
    for (index=0; index < NUM_TAD_REGS; index++)
    {
        tadReg = uncore_reg_read(TAD_RULE, index);

        tadLimit = ((uint64_t)tadReg << TAD_RULE_LIMIT_SHIFT) & ADDR_BIT_45_26_MSK;

        if ( (addr & ADDR_BIT_45_26_MSK) <= tadLimit)
        {
            /* We have found the correct TAD region for the given address */
            /* Abs value of Bits 9..8 + 1*/
            wayness = ((tadReg >> 8) & 0x00000003) + 1;
            break;
        }
    }

    if (index == NUM_TAD_REGS)
    {
        /* Only get here when a bad address was given */
        return -1;
    }

    chInterleave = (addr >> shiftCnt) & 0x03;

    /* If the mcchan hash is enabled, then update the chInterleave */
    if (sysDefeature2 & MCCHAN_HASH_ENABLE)
    {
        /* Shift the address down to the first bits which need to be XORed with chInterleave */
        addr >>= 12;

        /* Now XOR in the lower 2 bits with chInterleave shifting the address by 2 */
        /* for 8 different iterations */
        for (index = 0; index < 8; index++)
        {
            chInterleave ^= (addr & 0x03);
            addr >>= 2;
        }
    }

    /* Now get the tad_ch_tgtx bits.   This is accomplished by */
    /* shifting the tad register (interleave % wayness) * 2 bits to the left */
    return (tadReg >> ((chInterleave % wayness) * 2)) & 0x03;
}

static bool
enable_patrol_scrub_broadwell
    (
    bool enable
    )
{
    uint32_t val;

    val = uncore_reg_read(SCRUBCTL, 0);
    printk(KERN_DEBUG "ECC: patrol scrub 0x%x: %s\n", val, enable ? "enable" : "disable");

    if (enable)
    {
        /* turn on patrol scrub */
        uncore_reg_write(SCRUBCTL, 0, val | MEM_CTRL_SCRUB_EN);
    }
    else
    {
        /* turn off patrol scrub */
        uncore_reg_write(SCRUBCTL, 0, val & ~MEM_CTRL_SCRUB_EN);
    }

    /* Return 1/0 based on previous state of scrub control */
    return !!(val & MEM_CTRL_SCRUB_EN);
}

/* ECC error injection routine. Injects a single or multi bit error at a given
 * machine address. It temporarily disables the HW patrol scrub so no latent
 * errors are detected during injection (and so we don't get unlucky and inject
 * an error where the patrol scan is checking). Caching is disabled to force a
 * write to go out to physical memory. */
static void
ecc_inject_broadwell
    (
    uint64_t maddr,
    uint8_t bit,
    uint8_t immediate
    )
{
    bool                    scrub_state;
    uint32_t                injAddr_hi;
    uint32_t                injAddr_lo;
    int                     chan_num;
    unsigned long           flags;
    uint32_t                data;
    volatile uint32_t       *vaddr;
    uint32_t                inj_extra;
    uint32_t                xor_msk0;
    uint32_t                xor_msk1;
    unsigned long           prot = 0;
    struct vm_area_struct   *vma;
    struct mm_struct        *mm = current->mm;

    chan_num = maddr_to_channel(maddr);

    injAddr_hi = ((uint32_t) (maddr >> 35)) | 0x800;
    injAddr_lo = (uint32_t) ((maddr >> 3) & 0xFFFFFFFFULL);

    printk(KERN_INFO "ECC: chan %d, injAddr_hi %x, injAddr_lo %x, bit %x \n",
           chan_num, injAddr_hi, injAddr_lo, bit);

    if (chan_num == -1)
    {
        printk(KERN_ERR "ECC: Unexpected channel number - returning\n");
        return;
    }

    if (bit > 71 && bit != ECC_MULTI_BIT)
    {
        printk(KERN_ERR "ECC: Invalid bit number - returning\n");
        return;
    }


    /* remap input physical address to kernel virtual address */
    vma = find_vma(mm, maddr);
    if (!vma)
    {
        printk(KERN_ERR "ECC: find_vma fail, physAddr=0x%llx\n", maddr);
        return;
    }

    prot = pgprot_val(vma->vm_page_prot);
    vaddr = (volatile uint32_t *)ioremap_prot(maddr, sizeof(uint32_t), prot);
    if (!vaddr)
    {
        printk(KERN_ERR "ECC: ioremap fail, physAddr=0x%llx\n", maddr);
        return;
    }
    printk(KERN_INFO "ECC-ioremap: vaddr = %p, maddr = 0x%llx\n", vaddr, maddr);

    /* Protect access past here - many of these items are shared
     * across all cores. */
    spin_lock(&error_inj_lock);

    data = *vaddr;

    /* Suspend patrol scrubs. */
    printk(KERN_INFO "ECC: disable scrub\n");
    scrub_state = enable_patrol_scrub_broadwell(0);

    if (bit == ECC_MULTI_BIT)
    {
        /* flip some data bits to simulate MB corruption */
        data ^= 0x04000200;

        inj_extra = 0x00000402;
        xor_msk0  = 0x00000001;
        xor_msk1  = 0x00000004;
    }
    else
    {
        inj_extra = bit / 4;

        /* Get the XOR mask for the nibble */
        xor_msk0 = 1 << (bit %4);
        /* Now shift the nibble mask for the appropriate nibble inside
         * the 64-byte burst. */
        xor_msk0 <<= (((uint32_t)maddr & 0x38) / 2);
        xor_msk1 = 0x00000000;
    }

    /* This processor require OR'ing in 0x3c000000 per the Intel documentation */
    /* It is ignored by earlier processors */
    uncore_reg_write(INJ_EXTRA, chan_num, inj_extra | 0x3c000000);
    uncore_reg_write(DEV0_XOR_MASK, chan_num, xor_msk0);
    uncore_reg_write(DEV1_XOR_MASK, chan_num, xor_msk1);
    printk(KERN_INFO "ECC: inj_extra %x, dev0_xor_mask %x, dev1_xor_mask %x \n",
           inj_extra | 0x3c000000, xor_msk0, xor_msk1);

    uncore_reg_write(ADDR_MASK_HI, chan_num, 0);
    uncore_reg_write(ADDR_MASK_LO, chan_num, 0x7);
    uncore_reg_write(ADDR_MATCH_LO, chan_num, injAddr_lo);

    local_irq_save(flags);

    /* Disable and flush cache */
    printk(KERN_INFO "ECC: disable cache\n");
    cache_enable(0);

    /* No more printing until cache is enabled */

    wbinvd();

    /* Write to ADDR_MATCH_HI to "arm" the injection */
    uncore_reg_write(ADDR_MATCH_HI, chan_num, injAddr_hi);
    (void)uncore_reg_read(ADDR_MATCH_HI, chan_num);   /* readback to ensure write completes */

    *vaddr = data;

    /* Flush and re-enable cache */
    wbinvd();
    cache_enable(1);

    local_irq_restore(flags);

    /* resume patrol scrubs */
    printk(KERN_INFO "ECC: restore scrub\n");
    (void)enable_patrol_scrub_broadwell(scrub_state);

    /* End exclusive access */
    spin_unlock(&error_inj_lock);

    if (immediate)
    {
        printk(KERN_INFO "ECC: read ECC error\n");
        (void)*vaddr;
    }

    printk(KERN_INFO "ECC: iounmap");
    iounmap((void __iomem *)vaddr);
}

static uint32_t
msr_misc_to_bitmask
    (
    uint64_t misc
    )
{
    return (uint32_t)((misc >> 14) & 0x000FFFF);
}

static uint8_t
msr_status_to_device
    (
    uint64_t status
    )
{
    return (uint8_t)((status >> 32) & 0x0000001F);
}

static uint8_t
msr_status_to_dimm
    (
    uint64_t status,
    const struct ecc_banks* mce_ecc_banks
    )
{
    int mem_channel = status & 0x0FULL;
    if (mem_channel < 4)
        return mce_ecc_banks->dimm[mem_channel];

    return DIMM_INDETERM;
}

static uint32_t
msr_status_to_corr_err_cnt
    (
    uint64_t status
    )
{
    /* Bits 52-38 are corr err cnt - These are valid even if there is a MB
     * error. Upper bit is sticky across rollovers, so we could use that as
     * overflow detection if desired. Clear this count by clearing the sts
     * register. */
    return (uint32_t)((status >> 38) & 0x7FFF);
}

/* For this processor, the channel corresponds to the dimm */
static int32_t
maddr_to_dimm_broadwell
    (
    uint64_t addr
    )
{
    return (int32_t)maddr_to_channel(addr);
}

static uint32_t
log_ecc_bank
    (
    uint32_t *err_event,
    int bank,
    bool enable_log
    )
{
    uint32_t return_sts = 0;
    int sec_bank;
    uint64_t status;
    uint64_t addr;
    uint64_t misc;
    const struct ecc_banks* mce_ecc_banks = get_mce_ecc_banks();
    int mem_channel;
    struct ecc_error ecc_err = {0};

    /* If mce_ecc_banks isn't set, then MCE init has not yet been performed
     * so return 'no error'. */
    if (!mce_ecc_banks)
    {
        mce_printk(KERN_ERR, "Unable to log ECC bank %d\n", bank);
        return return_sts;
    }

    /* Read status and see if it is valid */
    rdmsrl(MSR_IA32_MCx_STATUS(bank), status);

    if (!(status & MCI_STATUS_VAL))
    {
        clear_bank_errors(bank);
        return return_sts;
    }

    if (status & MCI_STATUS_UC)
    {
        return_sts |= ECC_UNCOR;
    }

    /* Status is valid, get other info */
    /* NOTE - On some platforms, when a MB error is detected via
     * patrol scrub it reports an addr value slightly higher or lower than expected.
     * The exact amount is specific to the processor and based on the interleaving
     * algorithm.
     * This can cause the maddr_to_dimm() calculation to return the wrong DIMM.
     * We are not able to fully work around this.  Instead, we will rely on
     * the memory channel to indicate which DIMM instead of the address.
     */
    rdmsrl(MSR_IA32_MCx_ADDR(bank), addr);
    rdmsrl(MSR_IA32_MCx_MISC(bank), misc);

    /* clear status */
    clear_bank_errors(bank);

    /* Print message to shell/log */
    print_mce_msg(bank, status, addr, misc);

    ecc_err.bank = bank;
    ecc_err.maddr = addr;
    ecc_err.status = status;
    ecc_err.misc = misc;

    /* NOTE: If addr isn't valid, then this is going to be a bogus DIMM number.
     * This has only been seen when ECC reporting is disabled and we found a MB
     * error, plus other SB errors. The MB error is reported properly w/an addr
     * but the SB errors have an addr of 0x0, causing them to report on DIMM 0
     * when they're really on other DIMMs. Nothing currently checks/reports
     * mixed errors in this sort of detail, so just leave it this way for now.
     * We could derive a secondary DIMM # from the MC bank # if needed.
     */
    ecc_err.is_eos = is_eos_mem(addr);
    ecc_err.is_rpa = is_rpa_mem(addr);
    /* Poll for sb errors only */
    ecc_err.is_sb_enabled = 0;
    ecc_err.is_mb_enabled = mce_mb_enabled_broadwell();

    if (return_sts & ECC_UNCOR)
    {
        ecc_err.dimm = maddr_to_dimm_broadwell(addr);

        /* If uncorrectable error, then also clear the other reporting bank information
         * but do not use that data
         */
        if (bank == mce_ecc_banks->home_agent)
        {
            /* The mem_channel array must point to the bank that the HW will report
             * for the given channel #. On Soyuz, we don't use HW channel 0, so HW
             * should only report 1-3. Arapaho only has channels 0-1, both of which
             * could be in use. */
            mem_channel = status & 0x000000000000000FULL;
            if (mem_channel < 4 && (mce_ecc_banks->mem_channel[mem_channel] != 0))
            {
                clear_bank_errors(mce_ecc_banks->mem_channel[mem_channel]);
                ecc_err.sec_bank = mce_ecc_banks->mem_channel[mem_channel];
            }
        }
        else
        {
            /* Due to an errata on some processors, we cannot rely on the address
             * being reported for the error when not from home agent.
             *      (only happens if found during patrol scrub).
             * So instead, we should use the information in the status register
             * to determine which DIMM was affected. The address is only off by 128 bytes
             * worst case so no need to worry about usage when determining if in RPA or EOS.
             */
            ecc_err.dimm = msr_status_to_dimm(status, mce_ecc_banks);
        }
    }
    /* Take care of correctable errors which are not reported thru
     * the home agent.  This will happen if there are more than one
     * errors reported and this bank was not the last... or it is
     * being reported via patrol scrub
     */
    else if (bank != mce_ecc_banks->home_agent)
    {
        /* Determine the DIMM strictly from the status register
         * Since correctable errors are polled, there can be a multitude of errors
         * and no accurate record of the faulting address(es)
         */
        ecc_err.dimm = msr_status_to_dimm(status, mce_ecc_banks);

        /* Only report correctable errors if there are no uncorrectable ones */
        ecc_err.corr_err_cnt = msr_status_to_corr_err_cnt(status);

        if (ecc_err.corr_err_cnt)
        {
            return_sts |= ECC_COR;
            ecc_err.corr_bitmask = msr_misc_to_bitmask(misc);
            ecc_err.err_device = msr_status_to_device(status);
        }
    }
    /* Take care of correctable errors where the home agent is involved */
    else
    {
        ecc_err.dimm = maddr_to_dimm_broadwell(addr);

        /* The mem_channel array must point to the bank that the HW will report
         * for the given channel #. On Soyuz, we don't use HW channel 0, so HW
         * should only report 1-3. Arapaho only has channels 0-1, both of which
         * could be in use. */
        mem_channel = status & 0x000000000000000FULL;
        if (mem_channel < 4 && (mce_ecc_banks->mem_channel[mem_channel] != 0))
        {
            sec_bank = mce_ecc_banks->mem_channel[mem_channel];

            /* Always read/print so we have as much debug as possible.
             * Note that we're overwriting the top-level status and misc
             * values with the secondary ones. This is by design - these
             * are used to calculate correctable error info later.
             */
            rdmsrl(MSR_IA32_MCx_STATUS(sec_bank), status);
            rdmsrl(MSR_IA32_MCx_ADDR(sec_bank), addr);
            rdmsrl(MSR_IA32_MCx_MISC(sec_bank), misc);

            /* clear status */
            clear_bank_errors(sec_bank);

            print_mce_msg(sec_bank, status, addr, misc);

            /* Fill out the remaining fields in the ECC struct */
            ecc_err.sec_bank = sec_bank;
            ecc_err.sec_addr = addr;
            ecc_err.sec_status = status;
            ecc_err.sec_misc = misc;
        }

        ecc_err.corr_err_cnt = msr_status_to_corr_err_cnt(status);
        if (ecc_err.corr_err_cnt)
        {
            return_sts |= ECC_COR;
            ecc_err.corr_bitmask = msr_misc_to_bitmask(misc);
            ecc_err.err_device = msr_status_to_device(status);
        }
    }

    /* If there was a valid status but that status did not report
     * any MB or SB errors, flag it as an unknown error. We never
     * expect this to happen, but if it does, then we want to know
     * about it.
     */
    if ((return_sts & (ECC_COR | ECC_UNCOR)) == 0)
    {
        return_sts |= MCE_UNKNOWN;
    }

    if (return_sts & (ECC_UNCOR | MCE_UNKNOWN))
    {
        /* For now, setting this to true for the 'unknown' case as well */
        ecc_err.is_uncorrectable = 1;
    }

    if (ecc_err.is_rpa)
    {
        return_sts |= ECC_IN_RPA;
    }

    print_mce_end_msg(return_sts);

    if (enable_log)
        *err_event = log_error(*err_event, ERROR_TYPE_ECC, &ecc_err);

    return return_sts;
}

static uint32_t
log_mce_bank
    (
    uint32_t *err_event,
    int bank,
    bool enable_log
    )
{
    const struct ecc_banks* mce_ecc_banks = get_mce_ecc_banks();
    struct mce_error mce_err = {0};
    uint64_t status;
    uint64_t addr;
    uint64_t misc;
    uint32_t return_sts = 0;
    int mem_channel;
    int mc_bank;

    /* If mce_ecc_banks isn't set, then MCE init has not yet been performed
     * so return 'no error'. */
    if (!mce_ecc_banks)
    {
        mce_printk(KERN_ERR, "Unable to log MCE bank %d\n", bank);
        return return_sts;
    }

    /* If this is the Home Agent or one of the Mem Channel banks,
     * call the ECC logging function instead. */
    if (bank == mce_ecc_banks->home_agent)
    {
        return log_ecc_bank(err_event, bank, enable_log);
    }

    for (mem_channel = 0; mem_channel < 4; mem_channel++)
    {
        mc_bank = mce_ecc_banks->mem_channel[mem_channel];

        if (mc_bank && (bank == mc_bank))
        {
            return log_ecc_bank(err_event, bank, enable_log);
        }
    }

    rdmsrl(MSR_IA32_MCx_STATUS(bank), status);

    if (!(status & MCI_STATUS_VAL))
    {
        return return_sts; /* this bank has no valid telemetry */
    }

    if (status & MCI_STATUS_UC)
    {
        return_sts |= MCE_UNCOR;
    }

    mce_err.corr_err_cnt = msr_status_to_corr_err_cnt(status);
    if (mce_err.corr_err_cnt)
    {
        return_sts |= MCE_COR;
    }

    if (status & MCI_STATUS_ADDRV)
    {
        rdmsrl(MSR_IA32_MCx_ADDR(bank), addr);
        mce_err.dimm = maddr_to_dimm_broadwell(addr);
        mce_err.is_eos = is_eos_mem(addr);
        mce_err.is_rpa = is_rpa_mem(addr);
    }
    else
    {
        addr = 0xFFFFFFFFFFFFFFFFULL;
        mce_err.dimm = addr;
        mce_err.is_eos = 0;
        mce_err.is_rpa = 0;
    }

    if (status & MCI_STATUS_MISCV)
    {
        rdmsrl(MSR_IA32_MCx_MISC(bank), misc);
    }
    else
    {
        misc = 0xFFFFFFFFFFFFFFFFULL;
    }

    print_mce_msg(bank, status, addr, misc);

    /* Poll for sb errors only */
    mce_err.is_sb_enabled = 0;
    mce_err.is_mb_enabled = mce_mb_enabled_broadwell();

    /* log error */
    mce_err.maddr = addr;
    mce_err.status = status;
    mce_err.misc = misc;
    mce_err.msr_bank = bank;

    clear_bank_errors(bank);

    if (return_sts & MCE_UNCOR)
    {
        mce_err.is_uncorrectable = 1;
    }

    print_mce_end_msg(return_sts);

    wmb();

    if (enable_log)
        *err_event = log_error(*err_event, ERROR_TYPE_MCE, &mce_err);

    return return_sts;
}

static uint32_t
check_for_errors_broadwell
    (
    bool enable_log
    )
{
    uint32_t return_sts = 0;
    int mc_bank;
    int i;
    uint32_t err_event = 0;
    const struct ecc_banks* mce_ecc_banks = get_mce_ecc_banks();
    static int no_error_cnt = 0;

    /* If mce_ecc_banks isn't set, then MCE init has not yet been performed
     * so return 'no error'. */
    if (!mce_ecc_banks)
    {
        if (no_error_cnt < 10)
        {
            mce_printk(KERN_ERR, "Unable to check for ECC/MCE yet\n");
            no_error_cnt++;
            if (no_error_cnt == 10)
                mce_printk(KERN_ERR, "Disabling ECC/MCE warning\n");
        }
        return return_sts;
    }

    /* NOTE: We don't check PCC or OVER bits in the MCI_STATUS reg.
     * These can provide more/different info that we may wish to log
     * in the persistent region. */

    /* Check Home Agent (HA) first */
    return_sts |= log_ecc_bank(&err_event, mce_ecc_banks->home_agent, enable_log);

    /* Check through all mem channels */
    for (i = 0; i < 4; i++)
    {
        mc_bank = mce_ecc_banks->mem_channel[i];

        if (mc_bank == 0)
            continue; /* invalid bank (no mem channel N on this platform) */

        /* Force new event */
        err_event = 0;

        return_sts |= log_ecc_bank(&err_event, mc_bank, enable_log);
    }

    /* Done with ECC error reporting - now scan for MCE.
     * Some of these will overlap w/the ECC banks, the function is smart
     * enough to handle that.
     */
    for (i = 0; i < get_mce_bank_cnt(); i++)
    {
        /* Force new error event as we parse for other MCE events so that they
         * are reported separately
         */
        err_event = 0;
        return_sts |= log_mce_bank(&err_event, i, enable_log);
    }

    /* Return non-zero if something bad/unknown happened */
    return return_sts;
}

static const struct ecc_banks*
get_mce_ecc_banks
    (
    void
    )
{
    static struct ecc_banks* mce_ecc_banks = 0;

    /* If we already determined which structure, then simply return it. */
    if (mce_ecc_banks)
        return mce_ecc_banks;

    mce_printk(KERN_INFO, "MCE: Using Broadwell-DE ECC banks\n");
    mce_ecc_banks = &broadwell_de_ecc_banks;

    return mce_ecc_banks;
}

/* Enable/disable MCE for multibit errors.
 * This assumes that the MCE init has already been done. */
static void
mce_enable_mb_broadwell
    (
    bool enable
    )
{
    uint64_t val;
    const struct ecc_banks* mce_ecc_banks = get_mce_ecc_banks();

    if (!mce_ecc_banks)
        return;

    if (enable)
    {
        val = 0xffffffffffffffffULL;
    }
    else
    {
        val = 0ULL;
    }

    wrmsrl(MSR_IA32_MCx_CTL(mce_ecc_banks->home_agent), val);
}

/* Check multibit MCE state */
static bool
mce_mb_enabled_broadwell
    (
    void
    )
{
    uint64_t val;
    const struct ecc_banks* mce_ecc_banks = get_mce_ecc_banks();
    if (!mce_ecc_banks)
        return 0;

    rdmsrl(MSR_IA32_MCx_CTL(mce_ecc_banks->home_agent), val);
    return (val > 0);
}

static void
init_broadwell
    (
    void
    )
{
    /* Disable patrol scrub until EOS enables it */
    enable_patrol_scrub_broadwell(false);
}

ECC_INTERFACE ecc_iface_broadwell =
{
    .enable_patrol_scrub = enable_patrol_scrub_broadwell,
    .ecc_inject          = ecc_inject_broadwell,
    .mce_mb_enabled      = mce_mb_enabled_broadwell,
    .mce_enable_mb       = mce_enable_mb_broadwell,
    .check_for_errors    = check_for_errors_broadwell,
    .maddr_to_dimm       = maddr_to_dimm_broadwell,
    .init                = init_broadwell
};
