/*******************************************************************************

   ecc_ops.c
   Copyright 2016-2019 NetApp, Inc. All Rights Reserved.

   ECC Operations accessed via ioctls defined by eseries_kernel_interface

*******************************************************************************/
#include <linux/atomic.h>
#include <linux/slab.h>
#include <linux/uaccess.h>

#include "ecc.h"
#include "eki_common.h"
#include "eseries_kernel_interface.h"
#include "logging.h"

static atomic_t retrieve_in_progress;
static err_entry errbuf[NUM_ERR_ENTRIES];


/*
 * Enable or disable MB ECC errors from generating a machine check.
 */
int
eki_ecc_enable_mb
    (
    unsigned long arg
    )
{
    int     ret;
    int     mb_enable;

    ret = copy_from_user(&mb_enable, (int*)arg, sizeof(int));
    if (ret)
    {
        printk(KERN_ERR "eki-ECC MB enable: can't copy data from user space, %d\n", ret);
        ret = -EFAULT;
        return ret;
    }
    mce_enable_mb((bool)mb_enable);
    return 0;
}

/*
 * Check for MB ECC errors from generating a machine check.
 */
int
eki_ecc_check
    (
    unsigned long arg
    )
{
    int     ret;
    struct  eki_ecc_check check_data;
    uint32_t check_sts;

    ret = copy_from_user(&check_data, (struct eki_ecc_check*)arg, sizeof(struct eki_ecc_check));
    if (ret)
    {
        printk(KERN_ERR "eki-ECC check: can't copy data from user space, %d\n", ret);
        ret = -EFAULT;
        return ret;
    }

    check_sts = check_for_errors(check_data.log_enabled);
    if (check_sts)
        check_data.err_detected = 1;
    else
        check_data.err_detected = 0;

    ret = copy_to_user((struct eki_ecc_check*)arg, &check_data, sizeof(struct eki_ecc_check));
    if (ret)
    {
        printk(KERN_ERR "eki-ECC check: can't copy data to user space, %d\n", ret);
        ret = -EFAULT;
        return ret;
    }

    return 0;
}

/* eseries_kernel_interface wrapper function for ecc injection */
int
eki_inject_ecc
    (
    unsigned long arg
    )
{
    int     ret;
    struct  eki_eccinject_data ecc_data;

    ret = copy_from_user(&ecc_data, (struct eki_eccinject_data*)arg, sizeof(struct eki_eccinject_data));
    if (ret)
    {
        printk(KERN_ERR "eki-ECC injection: can't copy data from user space, %d\n", ret);
        ret = -EFAULT;
        return ret;
    }

    ecc_inject(ecc_data.physAddr, ecc_data.bit, ecc_data.immediate);
    return 0;
}

/* eseries_kernel_interface wrapper function for enable/disable ECC patrol scrub  */
int
eki_enable_scrub
    (
    unsigned long arg
    )
{
    int     ret;
    int     patrol_enable;

    ret = copy_from_user(&patrol_enable, (int*)arg, sizeof(int));
    if (ret)
    {
        printk(KERN_ERR "eki-ECC patrol scrub: can't copy data from user space, %d\n", ret);
        ret = -EFAULT;
        return ret;
    }
    enable_patrol_scrub((bool)patrol_enable);
    return 0;
}

/*
 * Converts address to DIMM
 */
int
eki_address_to_dimm
    (
    unsigned long arg
    )
{
    int ret;
    struct eki_addr_to_dimm info;

    ret = copy_from_user(&info, (int*)arg, sizeof(info));
    if (ret)
    {
        printk(KERN_ERR "%s: can't copy data from user space, %d\n", __func__, ret);
        ret = -EFAULT;
        return ret;
    }
    info.dimm = maddr_to_dimm(info.addr);

    ret = copy_to_user((struct eki_addr_to_dimm*)arg, &info, sizeof(info));
    if (ret)
    {
        printk(KERN_ERR "%s: can't copy data to user space, %d\n", __func__, ret);
        ret = -EFAULT;
        return ret;
    }
    return 0;
}

/*
 * Retrieve the error log entries
 */
int
eki_error_log_get
    (
    unsigned long arg
    )
{
    int      ret;
    uint32_t num_entries;
    struct   eki_log_data ecc_log;

    ret = copy_from_user(&ecc_log, (struct eki_log_data*)arg, sizeof(struct eki_log_data));
    if (ret)
    {
        printk(KERN_ERR "eki-ECC log get: can't copy data from user space, %d\n", ret);
        ret = -EFAULT;
        return ret;
    }

    if (atomic_add_unless(&retrieve_in_progress, 1, 1) == 0)
    {
        printk(KERN_ERR "eki-ECC log get: retrieve in progress %d\n", retrieve_in_progress.counter);
        ret = -EFAULT;
        return ret;
    }

    printk(KERN_DEBUG "ECC: call eki_error_log_get with %d\n", ecc_log.all_entries);
    num_entries = retrieve_errors(errbuf, ecc_log.all_entries);

    ret = copy_to_user((err_entry*)ecc_log.buffer, errbuf, num_entries * sizeof(err_entry));
    retrieve_in_progress.counter = 0;
    if (ret)
    {
        printk(KERN_ERR "eki-ECC log get: can't copy buffer to user space, %d\n", ret);
        ret = -EFAULT;
        return ret;
    }

    ecc_log.nr_entries = num_entries;
    ret = copy_to_user((struct eki_log_data*)arg, &ecc_log, sizeof(struct eki_log_data));
    if (ret)
    {
        printk(KERN_ERR "eki-ECC log get: can't copy num_entries to user space, %d\n", ret);
        ret = -EFAULT;
        return ret;
    }

    return 0;
}

/*
 * Acknowledge the error log entries
 */
int
eki_error_log_ack
    (
    unsigned long arg
    )
{
    int      ret;
    uint32_t nr_ackd;
    struct   eki_log_ack ecc_ack;

    ret = copy_from_user(&ecc_ack, (struct eki_log_ack*)arg, sizeof(struct eki_log_ack));
    if (ret)
    {
        printk(KERN_ERR "eki-ECC log ack: can't copy data from user space, %d\n", ret);
        ret = -EFAULT;
        return ret;
    }

    printk(KERN_DEBUG "ECC: call eki_error_log_ack with %d\n", ecc_ack.id);
    nr_ackd = acknowledge_error_event(ecc_ack.id);

    ecc_ack.nr_acked = nr_ackd;
    ret = copy_to_user((struct eki_log_ack*)arg, &ecc_ack, sizeof(struct eki_log_ack));
    if (ret)
    {
        printk(KERN_ERR "eki-ECC log ack: can't copy data to user space, %d\n", ret);
        ret = -EFAULT;
        return ret;
    }

    return 0;
}

/*
 * Dump the persistent error region to the journal
 */
void
eki_error_log_dump
    (
    void
    )
{
    printk(KERN_INFO "ECC: call eki_dump_log\n");
    dump_error_region();
}

/*
 * Insert error log event into the error log (for testing purposes)
 */
int
eki_error_log_add_event
    (
    unsigned long arg
    )
{
    int      ret;
    err_entry entry;

    ret = copy_from_user(&entry, (err_entry*)arg, sizeof(err_entry));
    if (ret)
    {
        printk(KERN_ERR "eki-ECC log add event: can't copy data from user space, %d\n", ret);
        ret = -EFAULT;
        return ret;
    }
    log_error(0, entry.type, &(entry.ecc));

    return 0;
}

/* ecc initial function */
int
eki_ecc_init
    (
    void
    )
{
    retrieve_in_progress.counter = 0;

    /* setup MCE handler and enable MCE and patrol scrub */
	return ecc_init();
}
