/*******************************************************************************

   ecc_skylake.c
   Copyright 2019 NetApp, Inc. All Rights Reserved.

   ECC handling for the skylake processor

   Supporting documentation:
      Intel Xeon Processor Scalable Memory Family
      EDS Volume 2 Part A Doc 546832

*******************************************************************************/
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/mm.h>
#include <linux/mm_types.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/uaccess.h>
#include <linux/string.h>
#include <asm/mce.h>
#include <asm/msr.h>
#include <asm/tlbflush.h>
#include <asm/paravirt.h>
#include <asm/page_types.h>

#include "ecc.h"
#include "eccP.h"
#include "logging.h"


/* Mask to get the memory channel from the status register */
#define STATUS_MCHAN_MASK   0x000000000000000FULL

enum imcRegister
{
    SCRUBCTL        = 1,
    PARITY_SYNDROME = 2,
    INJ_EXTRA       = 3,
    DEV0_XOR_MASK   = 4,
    DEV1_XOR_MASK   = 5,
    ADDR_MASK_HI    = 6,
    ADDR_MASK_LO    = 7,
    ADDR_MATCH_HI   = 8,
    ADDR_MATCH_LO   = 9
};

/* Number of IMC buses on this processor */
#define NUM_IMC_BUSES   2

/* Number of active channels for this processor */
#define  NUM_CHANNELS  2
typedef struct
{
    unsigned int imcBus;
    int m2m;
    int memChannel[NUM_CHANNELS];
    int dimm[NUM_CHANNELS];
} EccAccessInfo;

typedef struct
{
    int m2m;
    int memChannel;
} DimmAccessInfo;

#define DIMM_INDETERM   0xFF

/* Accessor functions to get various fields from the DRAM RULE register */
#define DRAM_RULE_ENABLE(reg)          ((reg >> 0)  & 0x1)
#define DRAM_INTERLEAVE_MODE(reg)      ((reg >> 1)  & 0x03)
#define DRAM_MOD3_AS_A_MOD2(reg)       ((reg >> 5)  & 0x03)
#define DRAM_LIMIT(reg)                ((reg >> 7)  & 0x0FFFFF)
#define DRAM_MOD3(reg)                 ((reg >> 27) & 0x01)
#define DRAM_MOD3_MODE(reg)            ((reg >> 30) & 0x03)


/* DATA DECLARATION */

static const EccAccessInfo eccAccessInfo[NUM_IMC_BUSES] =
{
    { 0x0a,        /* IMC bus */
      7,           /* M2M MSR */
        {13, 14},  /* Memory channels */
        {0,  1},   /* Corresponding DIMM identifier */
    },
    { 0x0c,        /* IMC bus */
      8,           /* M2M MSR */
        {15, 16},  /* Memory channels */
        {2,  3},   /* Corresponding DIMM identifiers */
    }
};

/* Bus containing the MCE/ECC registers */
static const unsigned int mceEccBus = 0x64;

static DEFINE_SPINLOCK(errorInjLock);

/* SCRUBCTL - SCRUB_EN bit */
static const uint32_t MEM_CTRL_SCRUB_EN = 0x80000000;

/* The interleave mode indicates which address bits specify the interleave mode.
 * So on reading the interleave mode, this array indicates how many bits to shift
 * right to get the actual interleave mode
 */
static const uint32_t interleaveModeToPackage[] = { 6, 8, 12, 30};

/* Converts Mod3_Mode register field to lower address bit definition for that mode */
static const uint32_t mod3ModeToSplitBit[] = { 6, 8, 12};

/* MSR bank for MLC (mid-level cache) */
static const int MLC_BANK = 3;

/***   FORWARD DECLARATIONS ***/

static uint32_t check_for_errors_skylake(bool enableLog);
static bool mce_mb_enabled_skylake(void);


/**
 * \brief  Command to read/write from IMC register
 *
 * \param[in]      bus      Which PCI bus
 * \param[in]      devFn    Which PCI device/function
 * \param[in]      offset   Offset into PCI config space
 * \param[in,out]  value    Pointer to value to read to/write from
 * \param[in]      write    Boolean indicating direction of operation (write=true, read=false)
 *
 * \return  0 if successful, non-zero if it failed
 */
static int
reg_rw
    (
    unsigned int bus,
    unsigned int devFn,
    unsigned int offset,
    uint32_t*    value,
    bool         write
    )
{
    int ret = 1;
    struct pci_dev* pdev = pci_get_domain_bus_and_slot(0, bus, devFn);
    if (pdev)
    {
        if (write)
        {
            ret = pci_write_config_dword(pdev, offset, *value);
        }
        else
        {
            ret = pci_read_config_dword(pdev, offset, value);
        }
    }
    return ret;
}

/**
 * \brief  Command to read/write from IMC register
 *
 * \param[in]      imcBusIdx      Which IMC Bus to read from /write to
 * \param[in]      regName        Which IMC register by name to read from/write to
 * \param[in]      channel        Where channel matters, indicates which channel
 * \param[in/out]  value          Pointer to value to read to/write from
 * \param[in]      write          Boolean indicating direction of operation (write=true, read=false)
 *
 * \return  0 if successful, non-zero if it failed
 */
static int
imc_reg_rw
    (
    unsigned int        imcBusIdx,
    enum imcRegister    regName,
    int                 channel,
    uint32_t*           value,
    bool                write
    )
{
    unsigned int dev = 0;
    unsigned int func = 0;
    unsigned int offset = 0;
    int ret = 1;

    if (imcBusIdx >= NUM_IMC_BUSES)
    {
        printk(KERN_ERR "%s: Invalid IMC device 0x%u:\n", __func__, imcBusIdx);
        return ret;
    }
    dev = eccAccessInfo[imcBusIdx].imcBus;

    if (channel >= NUM_CHANNELS)
    {
        printk(KERN_ERR "%s: Invalid IMC channel 0x%u:\n", __func__, channel);
        return ret;
    }

    switch (regName)
    {
        case SCRUBCTL:
            func = 0;
            offset = 0x914;
            break;

        case PARITY_SYNDROME:
            func = (channel == 0) ? 3 : 7;
            offset = 0x150;
            break;

        case INJ_EXTRA:
            func = (channel == 0) ? 3 : 7;
            offset = 0x208;
            break;

        case DEV0_XOR_MASK:
            func = (channel == 0) ? 3 : 7;
            offset = 0x200;
            break;

        case DEV1_XOR_MASK:
            func = (channel == 0) ? 3 : 7;
            offset = 0x204;
            break;

        case ADDR_MASK_HI:
            func = (channel == 0) ? 0 : 4;
            offset = 0x1CC;
            break;

        case ADDR_MASK_LO:
            func = (channel == 0) ? 0 : 4;
            offset = 0x1C8;
            break;

        case ADDR_MATCH_HI:
            func = (channel == 0) ? 0 : 4;
            offset = 0x1C4;
            break;

        case ADDR_MATCH_LO:
            func = (channel == 0) ? 0 : 4;
            offset = 0x1C0;
            break;

        default:
            printk(KERN_ERR "%s: Unknown IMC register request: %d\n", __func__, regName);
            return ret;
            break;
    }
    return reg_rw(mceEccBus, PCI_DEVFN(dev, func), offset, value, write);
}

/**
 * \brief  Command to read from IMC register
 *
 * \param[in]      imcBusIdx      Which IMC Bus to read from
 * \param[in]      regName        Which IMC register by name to read from
 * \param[in]      channel        Where channel matters, indicates which channel
 * \param[in,out]  value          Pointer to value to read to
 *
 * \return  0 if successful, non-zero if it failed
 */
static int
imc_reg_read
    (
    unsigned int        imcBusIdx,
    enum imcRegister    regName,
    int                 channel,
    uint32_t*           value
    )
{
    return imc_reg_rw(imcBusIdx, regName, channel, value, false);
}

/**
 * \brief  Command to write to IMC register
 *
 * \param[in]  imcBusIdx      Which IMC Bus to write to
 * \param[in]  regName        Which IMC register by name to write to
 * \param[in]  channel        Where channel matters, indicates which channel
 * \param[in]  value          Value to write if write operation specified
 *
 * \return  0 if successful, non-zero if it failed
 */
static int
imc_reg_write
    (
    unsigned int        imcBusIdx,
    enum imcRegister    regName,
    int                 channel,
    uint32_t            value
    )
{
    return imc_reg_rw(imcBusIdx, regName, channel, &value, true);
}

/**
 * \brief  Command to convert address to m2m device and channel
 *
 * \param[in]      addr   Address on which we are converting to dimm number
 * \param[in,out]  info   Pointer to structure which returns m2m/channel
 *
 * \return  Bool  indicating when true that the data is valid
 */
static bool
maddr_to_dimm_info
    (
    uint64_t addr,
    DimmAccessInfo* info
    )
{
    /* Maximum number of DRAM_RULE/INTERLEAVE_LIST pairs */
    static const int MAX_SAD_IDX = 20;

    /* Number of bits to shift to move limit to correct address location */
    static const int SAD_LIMIT_SHIFT = 26;

    static const uint32_t PACKAGE_MASK = 0x07;

    /* Bus/device/function for the SAD_ALL CHA */
    static const unsigned int SKL_BUS_1 = 0x16;
    static const unsigned int SAD_ALL = PCI_DEVFN(29, 0);
    static const unsigned int SAD = PCI_DEVFN(14, 0);
    static const unsigned int DRAM_RULE_OFFSET = 0x60;
    static const unsigned int INTERLEAVE_LIST_OFFSET = 0x64;
    static const unsigned int MC_ROUTE_TABLE_OFFSET = 0xB4;

    static const uint32_t INTERLEAVE_MASK_4BITS = 0x0F;
    static const uint32_t INTERLEAVE_MASK_3BITS = 0x07;
    static const uint32_t INTERLEAVE_MASK_BIT4  = 0x08;

    /* Invalid value for MOD3_mode */
    static const uint32_t MOD3MODE_INVALID = 3;

    static const uint32_t MAX_LOG_CH_INDEX = 5;

    static const uint32_t MC_ROUTE_TABLE_RING_BITS = 3;
    static const uint32_t MC_ROUTE_TABLE_RING_MASK = 0x07;
    static const uint32_t MC_ROUTE_TABLE_CHANNEL_ID_BITS = 2;
    static const uint32_t MC_ROUTE_TABLE_CHANNEL_ID_MASK = 0x03;
    static const uint32_t MC_ROUTE_TABLE_CHANNEL_ID_OFFSET = 18;

    int sadIdx;
    uint32_t interleaveMode;
    uint32_t chMod;
    uint32_t dramRule;
    uint32_t packageNum;
    uint32_t interleaveList;
    uint32_t mcRouteTable;
    uint64_t previousLimit = 0;
    uint64_t sLimit;
    uint32_t logChIndex;
    uint32_t tempLogChIndex;
    uint32_t addressMod;

    if (!info)
    {
        printk(KERN_ERR "%s: Pointer \"info\" uninitialized\n", __func__);
        return false;
    }

    if (reg_rw(SKL_BUS_1, SAD, MC_ROUTE_TABLE_OFFSET, &mcRouteTable, false))
    {
        printk(KERN_ERR "%s: Failure reading MC_ROUTE_TABLE register\n", __func__);
        return false;
    }

    /* The following code was derived from Intels cscripts.  Actual descriptions will be sparse. */
    for (sadIdx = 0; sadIdx < MAX_SAD_IDX; ++sadIdx)
    {
        if (reg_rw(SKL_BUS_1, SAD_ALL, DRAM_RULE_OFFSET + (sadIdx * 8), &dramRule, false))
        {
            printk(KERN_ERR "%s: Failure reading DRAM_RULE[%d] register\n", __func__, sadIdx);
            return false;
        }
        if (reg_rw(SKL_BUS_1, SAD_ALL, INTERLEAVE_LIST_OFFSET + (sadIdx * 8), &interleaveList, false))
        {
            printk(KERN_ERR "%s: Failure reading INTERLEAVE_LIST_CFG[%d] register\n", __func__, sadIdx);
            return false;
        }
        /* SAD limit is calculated by taking the register value, adding one to it, shifting
         * up to bits 45:26, then subtracting one.  The resultant value is the max address
         * for this SAD range.  This value will be valid even if the rule is not enabled.
         */
        sLimit = ((uint64_t)(DRAM_LIMIT(dramRule) + 1) << SAD_LIMIT_SHIFT) - 1;
        if (sLimit == previousLimit)
        {
            /* If the limit is same as the previous, then there are no more SAD definitions */
            /* so break since there is nothing left to check */
            break;
        }

        /* If the rule is not enabled, then simply set the previousLimit to this limit since
         * the limit value will always be valid
         */
        if (!DRAM_RULE_ENABLE(dramRule))
        {
            previousLimit = sLimit;
            continue;
        }

        /* If address is not in the limit range, then set previousLimit to sLimit and continue to next */
        if ((addr <= previousLimit) || (addr > sLimit))
        {
            previousLimit = sLimit;
            continue;
        }

        interleaveMode = DRAM_INTERLEAVE_MODE(dramRule);
        packageNum = (addr >> interleaveModeToPackage[interleaveMode]) & PACKAGE_MASK;

        /* If the MSB of the 4 bit interleave log channel is not set, then this is an
         * invalid configuration (remote memory).   Report that and return false.
         */
        logChIndex = (interleaveList >> (packageNum * 4)) & INTERLEAVE_MASK_4BITS;
        if ((logChIndex & INTERLEAVE_MASK_BIT4) == 0)
        {
            printk(KERN_ERR "%s: Failure reading INTERLEAVE_LIST_CFG[%d] register\n", __func__, sadIdx);
            return false;
        }
        logChIndex &= INTERLEAVE_MASK_3BITS;

        if (DRAM_MOD3(dramRule))
        {
            chMod = DRAM_MOD3_AS_A_MOD2(dramRule) ? 2 : 3;

            if (DRAM_MOD3_MODE(dramRule) == MOD3MODE_INVALID)
            {
                printk(KERN_ERR "%s: Invalid Mod3_mode (%d)\n", __func__, DRAM_MOD3_MODE(dramRule));
                return false;
            }
            addressMod = (uint32_t)((addr >> mod3ModeToSplitBit[DRAM_MOD3_MODE(dramRule)]) % chMod);
            if ((DRAM_MOD3_AS_A_MOD2(dramRule) == 0) || (DRAM_MOD3_AS_A_MOD2(dramRule) == 1))
                tempLogChIndex = (addressMod << 1);
            else
            {
                tempLogChIndex = (addressMod << 2);
                if (DRAM_MOD3_AS_A_MOD2(dramRule) == 2)
                    tempLogChIndex |= (addressMod ^ 1) << 1;
            }
            logChIndex = tempLogChIndex | (logChIndex & 1);
        }
        if (logChIndex > MAX_LOG_CH_INDEX)
        {
            printk(KERN_ERR "%s: Invalid logChIndex (%d)\n", __func__, logChIndex);
            return false;
        }

        info->m2m = (mcRouteTable >> (logChIndex * MC_ROUTE_TABLE_RING_BITS)) & MC_ROUTE_TABLE_RING_MASK;
        info->memChannel = (mcRouteTable >> ((logChIndex * MC_ROUTE_TABLE_CHANNEL_ID_BITS) +
                                              MC_ROUTE_TABLE_CHANNEL_ID_OFFSET)) & MC_ROUTE_TABLE_CHANNEL_ID_MASK;
        return true;

    }
    return false;
}

/**
 * \brief  Command to convert address to dimm number
 *
 * \param[in]  addr   Address on which we are converting to dimm number
 *
 * \return  dimm number
 */
static int32_t
maddr_to_dimm_skylake
    (
    uint64_t addr
    )
{
    DimmAccessInfo info;

    if (maddr_to_dimm_info(addr, &info))
    {
        if (info.m2m < NUM_IMC_BUSES && info.memChannel < NUM_CHANNELS)
            return eccAccessInfo[info.m2m].dimm[info.memChannel];
    }
    printk(KERN_ERR "%llx: Could not determine DIMM info\n", addr);
    return -1;
}

/**
 * \brief  Command to retrieve the correctable error count from status register
 *
 * \param[in]  status   value from status register
 *
 * \return  Correctable error count
 */
static uint32_t
msr_status_to_corr_err_cnt
    (
    uint64_t status
    )
{
    /* Bits 52-38 are corr err cnt - These are valid even if there is a MB
     * error. Upper bit is sticky across rollovers, so we could use that as
     * overflow detection if desired. Clear this count by clearing the status
     * register. */
    return (uint32_t)((status >> 38) & 0x7FFF);
}

/**
 * \brief  Command to convert MSR status into device number
 *
 * \param[in]  status   value from M2M status register
 *
 * \return  device number
 */
static uint8_t
msr_status_to_device
    (
    uint64_t status
    )
{
    return (uint8_t)((status >> 32) & 0x0000001F);
}

/**
 * \brief  Command to convert MSR status into DIMM number
 *
 * \param[in]      status    value from status register
 * \param[in,out]  eccAccess Pointer to structure containing ECC access info
 *
 * \return  DIMM number
 */
static uint8_t
msr_status_to_dimm
    (
    uint64_t status,
    const EccAccessInfo* eccAccess
    )
{
    int memChannel = status & STATUS_MCHAN_MASK;
    if (memChannel < NUM_CHANNELS)
        return eccAccess->dimm[memChannel];

    return DIMM_INDETERM;
}



/**
 * \brief  Command to enable/disable patrol scrub
 *
 * \param[in]  enable   Boolean indicate whether to enable or disable patrol scrub
 *
 * \return true patrol scrub was previously enabled and false otherwise
 */
static bool
enable_patrol_scrub_skylake
    (
    bool enable
    )
{
    uint32_t val = 0;
    int i;

    for (i = 0; i < NUM_IMC_BUSES; i++)
    {
        if (imc_reg_read(i, SCRUBCTL, 0, &val))
        {
            printk(KERN_ERR "%s: could not read from SCRUBCTL for IMC bus %d\n", __func__, eccAccessInfo[i].imcBus);
            continue;
        }
        printk(KERN_DEBUG "ECC - IMC Bus %d: patrol scrub 0x%x: %s\n",
            eccAccessInfo[i].imcBus, val, enable ? "enable" : "disable");

        if (enable)
        {
            /* turn on patrol scrub */
            if (imc_reg_write(i, SCRUBCTL, 0, val | MEM_CTRL_SCRUB_EN))
            {
                printk(KERN_ERR "%s: could not enable patrol scrub for IMC bus %d\n", __func__, eccAccessInfo[i].imcBus);
            }
        }
        else
        {
            /* turn off patrol scrub */
            if (imc_reg_write(i, SCRUBCTL, 0, val & ~MEM_CTRL_SCRUB_EN))
            {
                printk(KERN_ERR "%s: could not disable patrol scrub for IMC bus %d\n", __func__, eccAccessInfo[i].imcBus);
            }
        }
    }

    /* Return 1/0 based on previous state of scrub control */
    /* It is assumed that all devices report the same value */
    return !!(val & MEM_CTRL_SCRUB_EN);
}

/**
 * \brief   ECC error injection routine. Injects a single or multi bit error at a given
 *          machine address. It temporarily disables the HW patrol scrub so no latent
 *          errors are detected during injection (and so we don't get unlucky and inject
 *          an error where the patrol scan is checking). Caching is disabled to force a
 *          write to go out to physical memory.
 *
 * \param[in]  maddr      Address where ECC error should be injected
 * \param[in]  bit        Which bit should get corrupted (ECC_MULTI_BIT for multibit errors)
 * \param[in]  immediate  Flag when set to trigger the error right after injecting condition
 */
static void
ecc_inject_skylake
    (
    uint64_t maddr,
    uint8_t bit,
    uint8_t immediate
    )
{
    bool                    scrubState;
    uint32_t                injAddrHi;
    uint32_t                injAddrLo;
    unsigned long           flags;
    uint32_t                data;
    volatile uint32_t       *vaddr;
    uint32_t                injExtra;
    uint32_t                xorMsk0;
    uint32_t                xorMsk1;
    uint32_t                tmpVar;
    unsigned long           prot = 0;
    struct vm_area_struct   *vma;
    struct mm_struct        *mm = current->mm;
    DimmAccessInfo          info;

    if (!maddr_to_dimm_info(maddr, &info))
    {
        printk(KERN_ERR "ECC: Cannot determine channel info for address %llx\n", maddr);
        return;
    }

    injAddrHi = ((uint32_t) (maddr >> 35)) | 0x800;
    injAddrLo = (uint32_t) ((maddr >> 3) & 0xFFFFFFFFULL);

    printk(KERN_INFO "ECC: imc: %d chan %d, injAddrHi %x, injAddrLo %x, bit %x \n",
           info.m2m, info.memChannel, injAddrHi, injAddrLo, bit);

    if (bit > 71 && bit != ECC_MULTI_BIT)
    {
        printk(KERN_ERR "ECC: Invalid bit number - returning\n");
        return;
    }

    /* remap input physical address to kernel virtual address */
    vma = find_vma(mm, maddr);
    if (!vma)
    {
        printk(KERN_ERR "ECC: find_vma fail, physAddr=0x%llx\n", maddr);
        return;
    }

    prot = pgprot_val(vma->vm_page_prot);
    vaddr = (volatile uint32_t *)ioremap_prot(maddr, sizeof(uint32_t), prot);
    if (!vaddr)
    {
        printk(KERN_ERR "ECC: ioremap fail, physAddr=0x%llx\n", maddr);
        return;
    }
    printk(KERN_INFO "ECC-ioremap: vaddr = %p, maddr = 0x%llx\n", vaddr, maddr);

    /* Protect access past here - many of these items are shared
     * across all cores. */
    spin_lock(&errorInjLock);

    data = *vaddr;

    /* Suspend patrol scrubs. */
    printk(KERN_INFO "ECC: disable scrub\n");
    scrubState = enable_patrol_scrub_skylake(0);

    if (bit == ECC_MULTI_BIT)
    {
        /* flip some data bits to simulate MB corruption */
        data ^= 0x04000200;

        injExtra = 0x00000402;
        xorMsk0  = 0x00000001;
        xorMsk1  = 0x00000004;
    }
    else
    {
        injExtra = bit / 4;

        /* Get the XOR mask for the nibble */
        xorMsk0 = 1 << (bit % 4);
        /* Now shift the nibble mask for the appropriate nibble inside
         * the 64-byte burst. */
        xorMsk0 <<= (((uint32_t)maddr & 0x38) / 2);
        xorMsk1 = 0x00000000;
    }

    /* This processor require OR'ing in 0x3c000000 per the Intel documentation */
    /* It is ignored by earlier processors */
    imc_reg_write(info.m2m, INJ_EXTRA, info.memChannel, injExtra | 0x3c000000);
    imc_reg_write(info.m2m, DEV0_XOR_MASK, info.memChannel, xorMsk0);
    imc_reg_write(info.m2m, DEV1_XOR_MASK, info.memChannel, xorMsk1);
    printk(KERN_INFO "ECC: inj_extra %x, dev0_xor_mask %x, dev1_xor_mask %x \n",
           injExtra | 0x3c000000, xorMsk0, xorMsk1);

    imc_reg_write(info.m2m, ADDR_MASK_HI, info.memChannel, 0);
    imc_reg_write(info.m2m, ADDR_MASK_LO, info.memChannel, 0x7);
    imc_reg_write(info.m2m, ADDR_MATCH_LO, info.memChannel, injAddrLo);

    local_irq_save(flags);

    /* Disable and flush cache */
    printk(KERN_INFO "ECC: disable cache\n");
    cache_enable(0);

    /* No more printing until cache is enabled */

    wbinvd();

    /* Write to ADDR_MATCH_HI to "arm" the injection */
    imc_reg_write(info.m2m, ADDR_MATCH_HI, info.memChannel, injAddrHi);
    (void)imc_reg_read(info.m2m, ADDR_MATCH_HI, info.memChannel, &tmpVar);   /* readback to ensure write completes */

    *vaddr = data;
    /* Flush and re-enable cache */
    wbinvd();
    cache_enable(1);

    local_irq_restore(flags);

    /* resume patrol scrubs */
    printk(KERN_INFO "ECC: restore scrub\n");
    (void)enable_patrol_scrub_skylake(scrubState);

    /* End exclusive access */
    spin_unlock(&errorInjLock);

    if (immediate)
    {
        printk(KERN_INFO "ECC: read ECC error\n");
        (void)*vaddr;
    }

    printk(KERN_INFO "ECC: iounmap");
    iounmap((void __iomem *)vaddr);
}

/**
 * \brief   Log ECC errors for specified bank
 *
 * \param[out] errEvent     Pointer to error event
 * \param[in]  bank         Which MSR bank to check
 * \param[in]  enableLog    Flag when true to log errors
 *
 * \return     Non-zero value if errors were detected
 */
static uint32_t
log_ecc_bank
    (
    uint32_t *errEvent,
    int bank,
    bool enableLog
    )
{
    struct ecc_error eccErr;
    int secBank;
    uint64_t status;
    uint64_t addr;
    uint64_t misc;
    int imcIdx;
    int memChannel = 0;
    const EccAccessInfo* eccAccess = 0;
    bool isM2m = false;
    uint32_t returnSts = 0;

    memset(&eccErr, 0, sizeof(eccErr));

    /* Set eccAccess to structure element if this is a ECC or M2M bank */
    for (imcIdx = 0; imcIdx < NUM_IMC_BUSES; ++imcIdx)
    {
        if (bank == eccAccessInfo[imcIdx].m2m)
        {
            isM2m = true;
            eccAccess = &eccAccessInfo[imcIdx];
            break;
        }
        /* Check the channels to find the bank */
        for (memChannel = 0; memChannel < NUM_CHANNELS; ++memChannel)
        {
            if (bank == eccAccessInfo[imcIdx].memChannel[memChannel])
            {
                eccAccess = &eccAccessInfo[imcIdx];
                break;
            }
        }
    }

    if (!eccAccess)
    {
        printk(KERN_WARNING "%s: Invalid MSR bank 0x%u for ECC processing:\n", __func__, bank);
        return returnSts;
    }

    /* Read status and see if it is valid */
    rdmsrl(MSR_IA32_MCx_STATUS(bank), status);

    if (!(status & MCI_STATUS_VAL))
    {
        clear_bank_errors(bank);
        return returnSts;
    }

    if (status & MCI_STATUS_UC)
    {
        returnSts |= ECC_UNCOR;
    }

    /* Status is valid, get other info */
    /* NOTE - On some platforms, when a MB error is detected via
     * patrol scrub it reports an addr value slightly higher or lower than expected.
     * The exact amount is specific to the processor and based on the interleaving
     * algorithm.
     * This can cause the maddr_to_dimm_skylake() calculation to return the wrong DIMM.
     * We are not able to fully work around this.  Instead, we will rely on
     * the memory channel to indicate which DIMM instead of the address.
     */
    rdmsrl(MSR_IA32_MCx_ADDR(bank), addr);
    rdmsrl(MSR_IA32_MCx_MISC(bank), misc);

    /* clear status */
    clear_bank_errors(bank);

    /* Print message to shell/log */
    print_mce_msg(bank, status, addr, misc);

    eccErr.bank = bank;
    eccErr.maddr = addr;
    eccErr.status = status;
    eccErr.misc = misc;

    /* NOTE: If addr isn't valid, then this is going to be a bogus DIMM number.
     * This has only been seen when ECC reporting is disabled and we found a MB
     * error, plus other SB errors. The MB error is reported properly w/an addr
     * but the SB errors have an addr of 0x0, causing them to report on DIMM 0
     * when they're really on other DIMMs. Nothing currently checks/reports
     * mixed errors in this sort of detail, so just leave it this way for now.
     * We could derive a secondary DIMM # from the MC bank # if needed.
     */
    eccErr.is_eos = is_eos_mem(addr);
    eccErr.is_rpa = is_rpa_mem(addr);
    /* Poll for sb errors only */
    eccErr.is_sb_enabled = 0;
    eccErr.is_mb_enabled = mce_mb_enabled_skylake();

    if (returnSts & ECC_UNCOR)
    {
        eccErr.dimm = maddr_to_dimm_skylake(addr);

        /* If uncorrectable error, then also clear the other reporting bank information
         * but do not use that data.
         */
        if (isM2m)
        {
            /* If isM2m is set, then this is the M2M bank so get the associated
             * channel and clear the errors */
            memChannel = status & STATUS_MCHAN_MASK;
            if (memChannel < NUM_CHANNELS)
            {
                eccErr.sec_bank = eccAccess->memChannel[memChannel];
                clear_bank_errors(eccErr.sec_bank);
            }
        }
        else
        {
            /* Due to an errata on some processors, we cannot rely on the address
             * being reported for the error when not from M2M.
             *      (only happens if found during patrol scrub).
             * So instead, we should use the information in the status register
             * to determine which DIMM was affected. The address is only off by 128 bytes
             * worst case so no need to worry about usage when determining if in RPA or EOS.
             */
            eccErr.dimm = msr_status_to_dimm(status, eccAccess);
        }
    }

    /* Take care of correctable errors which are not reported thru
     * the M2M.  This will happen if there are more than one
     * errors reported and this bank was not the last... or it is
     * being reported via patrol scrub
     */
    else if (!isM2m)
    {
        /* Determine the DIMM strictly from the status register
         * Since correctable errors are polled, there can be a multitude of errors
         * and no accurate record of the faulting address(es)
         */
        eccErr.dimm = msr_status_to_dimm(status, eccAccess);

        /* Only report correctable errors if there are no uncorrectable ones */
        eccErr.corr_err_cnt = msr_status_to_corr_err_cnt(status);

        if (eccErr.corr_err_cnt)
        {
            returnSts |= ECC_COR;
            imc_reg_read(imcIdx, PARITY_SYNDROME, memChannel, &eccErr.corr_bitmask);
            /* bundle together the IMC index and the device number.   The
             * IMC index will be the upper 2 bits.   Device is a 5 bit value.
             */
            eccErr.err_device = (imcIdx << 6) | msr_status_to_device(status);
        }
    }
    /* Take care of correctable errors where the M2M is involved */
    else
    {
        eccErr.dimm = maddr_to_dimm_skylake(addr);

        /* The memChannel array must point to the bank that the HW will report
         * for the given channel #. */
        memChannel = status & STATUS_MCHAN_MASK;
        if (memChannel < NUM_CHANNELS)
        {
            secBank = eccAccess->memChannel[memChannel];

            /* Always read/print so we have as much debug as possible.
             * Note that we're overwriting the top-level status and misc
             * values with the secondary ones. This is by design - these
             * are used to calculate correctable error info later.
             */
            rdmsrl(MSR_IA32_MCx_STATUS(secBank), status);
            rdmsrl(MSR_IA32_MCx_ADDR(secBank), addr);
            rdmsrl(MSR_IA32_MCx_MISC(secBank), misc);

            /* clear status */
            clear_bank_errors(secBank);

            print_mce_msg(secBank, status, addr, misc);

            /* Fill out the remaining fields in the ECC struct */
            eccErr.sec_bank = secBank;
            eccErr.sec_addr = addr;
            eccErr.sec_status = status;
            eccErr.sec_misc = misc;
        }
        else  /* Make sure it is at least a value channel */
            memChannel = 0;

        eccErr.corr_err_cnt = msr_status_to_corr_err_cnt(eccErr.status);
        if (eccErr.corr_err_cnt)
        {
            returnSts |= ECC_COR;
            imc_reg_read(imcIdx, PARITY_SYNDROME, memChannel, &eccErr.corr_bitmask);
            /* bundle together the IMC index and the device number.   The
             * IMC index will be the upper 2 bits.   Device is a 5 bit value.
             */
            eccErr.err_device = (imcIdx << 6) | msr_status_to_device(eccErr.status);
        }
    }

    /* If there was a valid status but that status did not report
     * any MB or SB errors, flag it as an unknown error. We never
     * expect this to happen, but if it does, then we want to know
     * about it.
     */
    if ((returnSts & (ECC_COR | ECC_UNCOR)) == 0)
    {
        returnSts |= MCE_UNKNOWN;
    }

    if (returnSts & (ECC_UNCOR | MCE_UNKNOWN))
    {
        /* For now, setting this to true for the 'unknown' case as well */
        eccErr.is_uncorrectable = 1;
    }

    if (eccErr.is_rpa)
    {
        returnSts |= ECC_IN_RPA;
    }

    print_mce_end_msg(returnSts);

    if (enableLog)
        *errEvent = log_error(*errEvent, ERROR_TYPE_ECC, &eccErr);

    return returnSts;
}

/**
 * \brief   Log MCE errors for specified bank
 *
 * \param[out] errEvent     Pointer to error event
 * \param[in]  bank         Which MSR bank to check
 * \param[in]  enableLog    Flag when true to log errors
 *
 * \return     Non-zero value if errors were detected
 */
static uint32_t
log_mce_bank
    (
    uint32_t *errEvent,
    int bank,
    bool enableLog
    )
{
    struct mce_error mceErr;
    uint64_t status;
    uint64_t addr;
    uint64_t misc;
    uint32_t returnSts = 0;
    int imcIdx;
    int memChannel;

    memset(&mceErr, 0, sizeof(mceErr));

    /* If this is the M2M or one of the Mem Channel banks,
     * call the ECC logging function instead. */
    for (imcIdx = 0; imcIdx < NUM_IMC_BUSES; ++imcIdx)
    {
        if (bank == eccAccessInfo[imcIdx].m2m)
        {
            return log_ecc_bank(errEvent, bank, enableLog);
        }

        for (memChannel = 0; memChannel < NUM_CHANNELS; ++memChannel)
        {
            if (bank == eccAccessInfo[imcIdx].memChannel[memChannel])
            {
                return log_ecc_bank(errEvent, bank, enableLog);
            }
        }
    }

    rdmsrl(MSR_IA32_MCx_STATUS(bank), status);

    if (!(status & MCI_STATUS_VAL))
    {
        return returnSts;              /* this bank has no valid telemetry */
    }

    if (status & MCI_STATUS_UC)
    {
        returnSts |= MCE_UNCOR;
    }

    mceErr.corr_err_cnt = msr_status_to_corr_err_cnt(status);
    if (mceErr.corr_err_cnt)
    {
        returnSts |= MCE_COR;
    }

    if (status & MCI_STATUS_ADDRV)
    {
        rdmsrl(MSR_IA32_MCx_ADDR(bank), addr);
        mceErr.dimm = maddr_to_dimm_skylake(addr);
        mceErr.is_eos = is_eos_mem(addr);
        mceErr.is_rpa = is_rpa_mem(addr);
    }
    else
    {
        addr = 0xFFFFFFFFFFFFFFFFULL;
        mceErr.dimm = addr;
        mceErr.is_eos = 0;
        mceErr.is_rpa = 0;
    }

    if (status & MCI_STATUS_MISCV)
    {
        rdmsrl(MSR_IA32_MCx_MISC(bank), misc);
    }
    else
    {
        misc = 0xFFFFFFFFFFFFFFFFULL;
    }

    print_mce_msg(bank, status, addr, misc);

    /* Poll for sb errors only */
    mceErr.is_sb_enabled = 0;
    mceErr.is_mb_enabled = mce_mb_enabled_skylake();

    /* log error */
    mceErr.maddr = addr;
    mceErr.status = status;
    mceErr.misc = misc;
    mceErr.msr_bank = bank;

    /* If this is a cache error reporting MSR, then set the is_cache_msr flag */
    mceErr.is_cache_msr = (bank == MLC_BANK);

    clear_bank_errors(bank);

    if (returnSts & MCE_UNCOR)
    {
        mceErr.is_uncorrectable = 1;
    }

    print_mce_end_msg(returnSts);

    wmb();

    if (enableLog)
        *errEvent = log_error(*errEvent, ERROR_TYPE_MCE, &mceErr);

    return returnSts;
}

/**
 * \brief   Wrapper function to call processor specific routine to check for errors
 *
 * \param[in]  enableLog     Enable writing to error log
 *
 * \return     Non-zero value if errors were detected
 */
static uint32_t
check_for_errors_skylake
    (
    bool enableLog
    )
{
    uint32_t returnSts = 0;
    uint32_t errEvent = 0;
    int imcIdx;
    int memChannel;
    int bankIdx;

    /* NOTE: We don't check PCC or OVER bits in the MCI_STATUS reg.
     * These can provide more/different info that we may wish to log
     * in the persistent region. */

    /* Check M2M banks first */
    for (imcIdx = 0; imcIdx < NUM_IMC_BUSES; ++imcIdx)
    {
        returnSts |= log_ecc_bank(&errEvent, eccAccessInfo[imcIdx].m2m, enableLog);

        /* Check through all mem channels */
        for (memChannel = 0; memChannel < NUM_CHANNELS; ++memChannel)
        {
            /* Force new event */
            errEvent = 0;

            returnSts |= log_ecc_bank(&errEvent, eccAccessInfo[imcIdx].memChannel[memChannel], enableLog);
        }
    }

    /* Done with ECC error reporting - now scan for MCE.
     * Some of these will overlap w/the ECC banks, the function is smart
     * enough to handle that.
     */
    for (bankIdx = 0; bankIdx < get_mce_bank_cnt(); ++bankIdx)
    {
        /* Force new error event as we parse for other MCE events so that they
         * are reported separately
         */
        errEvent = 0;
        returnSts |= log_mce_bank(&errEvent, bankIdx, enableLog);
    }

    /* Return non-zero if something bad/unknown happened */
    return returnSts;
}

/* Enable/disable MCE for multibit errors.
 * This assumes that the MCE init has already been done. */
static void
mce_enable_mb_skylake
    (
    bool enable
    )
{
    int i;
    for (i = 0; i < NUM_IMC_BUSES; ++i)
    {
        wrmsrl(MSR_IA32_MCx_CTL(eccAccessInfo[i].m2m), enable ?  0xffffffffffffffffULL : 0ULL);
    }
    /* Also need to disable mid-level cache bank or it will report errors when ECC banks are disabled */
    wrmsrl(MSR_IA32_MCx_CTL(MLC_BANK), enable ?  0xffffffffffffffffULL : 0ULL);
}

/* Check multibit MCE state */
static bool
mce_mb_enabled_skylake
    (
    void
    )
{
    bool result = false;
    uint64_t val;
    int i;
    for (i = 0; i < NUM_IMC_BUSES; ++i)
    {
        rdmsrl(MSR_IA32_MCx_CTL(eccAccessInfo[i].m2m), val);
        result |= (val != 0);
    }
    return result;
}

static void
init_skylake
    (
    void
    )
{
    /* Disable patrol scrub until EOS enables it */
    enable_patrol_scrub_skylake(false);
}

ECC_INTERFACE ecc_iface_skylake =
{
    .enable_patrol_scrub = enable_patrol_scrub_skylake,
    .ecc_inject          = ecc_inject_skylake,
    .mce_mb_enabled      = mce_mb_enabled_skylake,
    .mce_enable_mb       = mce_enable_mb_skylake,
    .check_for_errors    = check_for_errors_skylake,
    .maddr_to_dimm       = maddr_to_dimm_skylake,
    .init                = init_skylake
};
