/*******************************************************************************

    eseries_kernel_interface.c
    Copyright 2016-2019 NetApp, Inc. All Rights Reserved.

    EKI driver module

*******************************************************************************/
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/pci.h>

#include "ecc_ops.h"
#include "misc_ops.h"
#include "nmihandler.h"
#include "eki_common.h"
#include "eseries_kernel_interface.h"
#include "persistent_region.h"
#include "rmem-mgr.h"

#define EKI_VERSION      "1.0"
#define BDW_DMI          0x6f00
#define SKL_DMI          0x2020

enum cpuModel cpu = CPU_UNKNOWN;

/* function declaration */
long eki_ioctl(struct file* filep, unsigned int cmd, unsigned long arg);
void cpuModelGet(void);

/* eki device driver data structure */
struct eki_dev
{
    dev_t                       devno;                  /* char device number */
    struct cdev                 cdev;                   /* a char device */
    struct class*               dev_class;              /* sysfs device class */
};

static struct eki_dev           ekiDev;

/* char dev supported file operations */
static struct file_operations eki_fops =
{
    .owner              = THIS_MODULE,
    .unlocked_ioctl     = eki_ioctl
};

/* eseries kernel interface char device IOCTL function
 *
 * It is used by user space to call ioctls
 */
long
eki_ioctl(struct file* filep, unsigned int cmd, unsigned long arg)
{
    int  ret = 0;

    switch (cmd)
    {
    case EKI_IOCTL_NMI_SEND_FDS:
        /* obtain eventfds for nmi handler */
        ret = nmi_pci_get_efds(arg);
        break;
    case EKI_IOCTL_MMCONFIG_GET:
        ret = eki_pci_get_mmconfig(arg);
        break;
    case EKI_IOCTL_ECC_ENABLE_MB:
        ret = eki_ecc_enable_mb(arg);
        break;
    case EKI_IOCTL_ECC_CHECK:
        ret = eki_ecc_check(arg);
        break;
    case EKI_IOCTL_ECC_INJECT:
        ret = eki_inject_ecc(arg);
        break;
    case EKI_IOCTL_ECC_SCRUB_ENABLE:
        ret = eki_enable_scrub(arg);
        break;
    case EKI_IOCTL_ADDR_TO_DIMM:
        ret = eki_address_to_dimm(arg);
        break;
    case EKI_IOCTL_LOG_GET:
        ret = eki_error_log_get(arg);
        break;
    case EKI_IOCTL_LOG_ACK:
        ret = eki_error_log_ack(arg);
        break;
    case EKI_IOCTL_LOG_DUMP:
        eki_error_log_dump();
        break;
    case EKI_IOCTL_LOG_EVENT:
        ret = eki_error_log_add_event(arg);
        break;
    case EKI_IOCTL_CPUFREQ_GET:
        ret = eki_cpufreq_get(arg);
        break;
    case EKI_IOCTL_REL_CPUTEMP_GET:
        ret = eki_relative_cputemp_get(arg);
        break;
    case EKI_IOCTL_TCCTEMP_GET:
        ret = eki_tcctemp_get(arg);
        break;
    case EKI_IOCTL_PCI_SHUTDOWN:
        ret = eki_pci_shutdown();
        break;
    case EKI_IOCTL_TSOD_DISABLE:
        ret = eki_tsod_disable(arg);
        break;
    case EKI_IOCTL_TSOD_RESTORE:
        ret = eki_tsod_restore(arg);
        break;
    case EKI_IOCTL_WB_INVALIDATE:
        wbinvd();
        break;
    case EKI_IOCTL_PER_REG_IMT:
        ret = eki_per_reg_imt(arg);
        break;
    default:
        printk(KERN_ERR "eki handler: received unknown argument: %u\n", cmd);
        return -EINVAL;
    }
    return ret;
}

/**
 * obtain CPU model routine
 *
 * obtain CPU model based on
 * device id
 */
void cpuModelGet(void)
{
    if (pci_get_device(PCI_VENDOR_ID_INTEL, BDW_DMI, NULL))
    {
        printk(KERN_INFO "NMI: Using Broadwell-DE processor\n");
        cpu = BROADWELL;
        return;
    }

    if (pci_get_device(PCI_VENDOR_ID_INTEL, SKL_DMI, NULL))
    {
        printk(KERN_INFO "NMI: Using Skylake-D processor\n");
        cpu = SKYLAKE;
        return;
    }

    printk(KERN_ERR "EKI: Unknown Intel chipset - unable to determine processor\n");
    cpu = CPU_UNKNOWN;
    return;
}

/* Module initialization function
 *
 * This routine will create a new char device under /dev
 * and a new class under /sys. It also register our NMI
 * handler for calling back when a NMI is issued.
 */
static int __init
eki_init(void)
{
    int                 ret = 0;
    struct device*      device;
    void* kAddr;
    unsigned long kLen;
    unsigned long kPhysAddr;

    /* get cpu model */
    cpuModelGet();

    /* ecc init */
    if (eki_ecc_init() != 0)
    {
        printk(KERN_ERR "Unable to initialize ecc error handling\n");
        return -1;
    }

    /* Initialize the persistent logging region */
    kAddr = rmem_kernel_log_memory(&kLen, &kPhysAddr);
    if (init_persistent_region(kAddr, kLen, kPhysAddr) != 0)
    {
        printk(KERN_ERR "Unable to initialize error logging region\n");
        return -1;
    }

    printk(KERN_INFO "Initializing eki char device ...\n");

    /* /proc/devices */
    ret = alloc_chrdev_region(&ekiDev.devno, 0, 1, EKI_DEV_NAME);
    if (ret)
    {
        printk(KERN_ERR "Unable to allocate device %s: %d\n", EKI_DEV_NAME, ret);
        return ret;
    }

    cdev_init(&ekiDev.cdev, &eki_fops);
    ekiDev.cdev.owner = THIS_MODULE;

    ret = cdev_add(&ekiDev.cdev, ekiDev.devno, 1);
    if (ret)
    {
        printk(KERN_ERR "Unable to add device %s: %d\n", EKI_DEV_NAME, ret);
        unregister_chrdev_region(ekiDev.devno, 1);
        return ret;
    }

    /* /sys/class */
    ekiDev.dev_class = class_create(THIS_MODULE, EKI_DEV_NAME);
    if (IS_ERR(ekiDev.dev_class))
    {
        ret = PTR_ERR(ekiDev.dev_class);
        printk(KERN_ERR "Unable to register device %s on sysfs: %d\n", EKI_DEV_NAME,ret);
        cdev_del(&ekiDev.cdev);
        unregister_chrdev_region(ekiDev.devno, 1);
        return ret;
    }

    device = device_create(ekiDev.dev_class, NULL, ekiDev.devno, NULL, EKI_DEV_NAME);
    if (IS_ERR(device))
    {
        ret = PTR_ERR(device);
        printk(KERN_ERR "Unable to create device %s on sysfs: %d\n", EKI_DEV_NAME, ret);
        cdev_del(&ekiDev.cdev);
        unregister_chrdev_region(ekiDev.devno, 1);
        class_destroy(ekiDev.dev_class);
        return ret;
    }

    /* register nmi handler */
    ret = nmi_pci_handler_registration(1);

    return ret;
}

/* Module exit function
 *
 * Unregister our nmi handler.
 * Delete eki class in sysfs.
 * Delete eki char device.
 */
static void __exit
eki_exit(void)
{
    /* unregister nmi pci handler */
    nmi_pci_handler_registration(0);

    printk(KERN_INFO "Unregistering eseries kernel interface char device from sysfs and proc\n");
    device_destroy(ekiDev.dev_class, ekiDev.devno);
    class_destroy(ekiDev.dev_class);
    cdev_del(&ekiDev.cdev);
    unregister_chrdev_region(ekiDev.devno, 1);
}

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("NetApp eseries kernel interface" EKI_VERSION);
MODULE_AUTHOR("NetApp, Inc");

module_init(eki_init);
module_exit(eki_exit);

