/*******************************************************************************

NAME            logging.c
SUMMARY         Error logging code

        Copyright 2016-2018 NetApp, Inc. All Rights Reserved.

*******************************************************************************/

#include <linux/bug.h>
#include <linux/string.h>
#include <linux/time.h>
#include <linux/timer.h>

#include "ecc.h"
#include "logging.h"
#include "persistent_region.h"

/*
 * base addr of pages for persistent error info region
 *
 * err_metadata should point to this location
 */
static void*         error_info_addr;
static err_metadata* metadata = 0;


static void initialize_region(void);

/*
 * Log an error entry in the error region.
 * If the specified event id is 0, a new event id will be returned.
 * Otherwise, the specified event id is used to relate multiple entries
 * across multiple calls to this function.
 *
 *  event_id - event id for this entry or 0 to determine next
 *  type - type of error (see error_ops.h)
 *  data - structure corresponding to type of error, actual data logged
 *
 * returns event_id used in the new entry
 */
uint32_t
log_error
    (
    uint32_t event_id,  /* 0 for 'next' */
    uint8_t type,
    void* data          /* cast to struct type indicated by previous argument */
    )
{
    err_entry*  entry;
    uint64_t    addr;
    int         i;
    struct timespec tv;

    if (metadata == 0)
    {
        mce_printk(KERN_ERR, "Error region not ready yet\n");
        return 0;
    }

    /*
     * If an ECC error occurred in the log region we need to
     * scrub and reinitialize the region before using it.
     */
    if (type == ERROR_TYPE_ECC)
    {
        addr = ((struct ecc_error*) data)->maddr;

        if (addr >= (uint64_t) persistent_region_phys_addr &&
           addr < (((uint64_t) persistent_region_phys_addr) + (persistent_region_len)) )
        {
            mce_printk(KERN_ERR, "ECC error in log region! Reinitializing...\n");
            init_persistent_region(persistent_region_addr,
                                   persistent_region_len,
                                   persistent_region_phys_addr);
        }
    }

    /* Look for the next unused entry */
    for(i=0; i < NUM_ERR_ENTRIES; i++)
    {
        entry = metadata->error + i;
        if (entry->state == ERROR_ENTRY_EMPTY)
            break;
    }

    /* Bail out if no available entries */
    if (entry->state != ERROR_ENTRY_EMPTY)
    {
        metadata->overflow++;
        mce_printk(KERN_ERR, "No unused entries found! (overflow=%d)\n", metadata->overflow);
        return 0;
    }

    /* Fill in new entry */
    entry->state = ERROR_ENTRY_WRITING;
    /* Using system time, not RTC */
    getnstimeofday(&tv);
    entry->timestamp = tv.tv_sec;
    entry->timestamp = 0;

    /* Continue writing entrys for a given event id or get the next available event id */
    if (event_id)
        entry->id = event_id;
    else
    {
        entry->id = metadata->next_id;

        if (metadata->next_id == 0xffffffff)
        {
            metadata->next_id = 1;
            mce_printk(KERN_ERR, "Error logging event id rolled over.\n");
        }
        else
            metadata->next_id++;
    }

    /* set type and type-specific data */
    entry->type = type;
    switch(type)
    {
        case ERROR_TYPE_MCE:
            memcpy(&entry->mce, data, sizeof(struct mce_error));
            break;
        case ERROR_TYPE_PCI:
            memcpy(&entry->pci, data, sizeof(struct pci_error));
            break;
        case ERROR_TYPE_REG:
            memcpy(&entry->reg, data, sizeof(struct reg_error));
            break;
        case ERROR_TYPE_ECC:
            memcpy(&entry->reg, data, sizeof(struct ecc_error));
            break;
        default:
            mce_printk(KERN_ERR, "Unknown error type %d for event %d\n", type, entry->id);
            break;
    }

    /* Mark entry as complete */
    entry->state = ERROR_ENTRY_USED;

    mce_printk(KERN_INFO, "Logged error: event_id=%d, type=%d\n", entry->id, type);

    return entry->id;
}

/*
 * Populate a given buffer with error entries.
 * Errors logged in the error region will be copied into the provided
 * buffer, grouped by event id.
 *
 *  errors - buffer to copy in error entries
 *      Note: must be at least as large as the error region!
 *            NUM_ERR_ENTRIES * sizeof(err_entry)
 *
 *  all - if set, will return all entries in the error region, otherwise only USED
 *        entries (those un-acknowledged) will be reported.
 *
 * returns number of error entries populated in the provided buffer
 */
uint32_t
retrieve_errors
    (
    err_entry* errors,
    bool all
    )
{
    uint32_t        copied = 0;
    int             event_idx = 0;
    int             found;
    int             i;
    int             j;
    err_entry*      entry;
    static uint32_t events[NUM_ERR_ENTRIES];

    if (metadata == 0)
    {
        printk(KERN_ERR "Error region not ready yet\n");
        return 0;
    }

    /* find all the events ids that we care about */
    memset(&events, 0, NUM_ERR_ENTRIES * sizeof(uint32_t));
    entry = metadata->error;
    for(i=0; i < NUM_ERR_ENTRIES; i++)
    {
        if (entry->id != events[i]
           && (entry->state == ERROR_ENTRY_USED || all) )
        {
            found = 0;
            for(j=0; j < event_idx; j++)
            {
                if (events[j] == entry->id)
                {
                    found = 1;
                    break;
                }
            }

            if (found == 0)
                events[event_idx++] = entry->id;
        }
        entry++;
    }

    /* fill error entry array, grouped by event id */
    for(i=0; i < event_idx; i++)
    {
        entry = metadata->error;
        for(j=0; j < NUM_ERR_ENTRIES; j++)
        {
            if (entry->id == events[i])
            {
                memcpy(errors + copied, entry, sizeof(err_entry));
                copied++;
            }
            entry++;
        }
    }

    printk(KERN_INFO "Retrieved %d error entries\n", copied);

    return copied;
}

/*
 * Mark all entries for a given event_id as are acknowledged.
 * This returns those entries to an UNUSED state.
 *
 *  event_id - event id to acknowledge
 *
 * returns number of error entries acknowledged
 */
uint32_t
acknowledge_error_event
    (
    uint32_t event_id
    )
{
    err_entry*  entry;
    uint32_t    nr_ackd = 0;
    int         i;

    if (metadata == 0)
    {
        printk(KERN_ERR "Error region not ready yet\n");
        return 0;
    }

    for(i=0; i < NUM_ERR_ENTRIES; i++)
    {
        entry = metadata->error + i;
        if (entry->id == event_id && entry->state != ERROR_ENTRY_EMPTY)
        {
            entry->state = ERROR_ENTRY_EMPTY;
            nr_ackd++;
        }
    }

    printk(KERN_INFO "Acknowledged %d entries for event_id=%d\n", nr_ackd, event_id);

    return nr_ackd;
}

/*
 * Debug utility to dump the error info metadata and entries to the journal
 */
void
dump_error_region
    (
    void
    )
{
    bool        headerDisplayed = 0;
    err_entry*  entry;
    int         i;

    if (metadata == 0)
    {
        printk(KERN_INFO "Error region not ready yet\n");
        return;
    }

    for(i=0; i < NUM_ERR_ENTRIES; i++)
    {
        entry = metadata->error + i;

        if (entry->state == ERROR_ENTRY_EMPTY)
            continue;

        // Don't display header unless errors found
        if (!headerDisplayed)
        {
            struct error_region* err_logging = (struct error_region*)error_info_addr;

            printk(KERN_INFO "Error signature: %s, version: %d\n",
                err_logging->signature, err_logging->version);
            printk(KERN_INFO "Error metadata:\n");
            printk(KERN_INFO "error*: %p, next_id: %d, overflow: %d\n",
                   metadata->error, metadata->next_id, metadata->overflow);

            printk(KERN_INFO "Error entries:\n");
            printk(KERN_INFO "## event state type timestamp        data\n");

            headerDisplayed = 1;
        }

        printk(KERN_INFO "%02d %05d %1d     %1d    %llu ",
                i, entry->id, entry->state, entry->type, entry->timestamp);

        switch(entry->type)
        {
            case ERROR_TYPE_MCE:
                printk(KERN_INFO "%llu %llu %llu %2d %1d %1d %1d %1d %1d",
                        entry->mce.maddr, entry->mce.status, entry->mce.misc,
                        entry->mce.msr_bank, entry->mce.dimm, entry->mce.is_sb_enabled,
                        entry->mce.is_mb_enabled, entry->mce.is_uncorrectable,
                        entry->mce.is_eos);
                break;
            case ERROR_TYPE_PCI:
                printk(KERN_INFO "%1d %02x:%02x:%02x %04x:%04x %04x %04x %04x %04x %04x %04x %04x",
                        entry->pci.pci_type, entry->pci.bus, entry->pci.device,
                        entry->pci.func, entry->pci.vid, entry->pci.did,
                        entry->pci.pci_status, entry->pci.pci_sec_status,
                        entry->pci.pcie_type, entry->pci.pcie_dev_status,
                        entry->pci.pcie_slot_status, entry->pci.pcie_link_status,
                        entry->pci.pcie_link_status2);
                break;
            case ERROR_TYPE_REG:
                printk(KERN_INFO "%08x %08x %08x %08x %08x %08x %08x",
                        entry->reg.gn_status, entry->reg.gf_status, entry->reg.gc_status,
                        entry->reg.irpp0_status, entry->reg.irpp1_status,
                        entry->reg.iio_status, entry->reg.mi_status);
                break;
            case ERROR_TYPE_ECC:
                printk(KERN_INFO "%llu %llu %llu %2d %1d %1d %1d %1d %1d",
                        entry->ecc.maddr, entry->ecc.status, entry->ecc.misc,
                        entry->ecc.bank, entry->ecc.dimm, entry->ecc.is_sb_enabled,
                        entry->ecc.is_mb_enabled, entry->ecc.is_uncorrectable,
                        entry->ecc.is_eos);
                break;
            default:
                break;
        }

        printk(KERN_INFO "\n");
    }
}

static void initialize_region
    (
    void
    )
{
    struct error_region* err_logging = (struct error_region*)error_info_addr;

    if (!metadata)
    {
        mce_printk(KERN_ERR, "Error: No pointer provided for initialize_region()\n");
        return;
    }

    /* initialize metadata */
    memcpy(err_logging->signature, errorRegionSignature, sizeof(err_logging->signature));
    err_logging->version = errorRegionVersion;

    printk(KERN_INFO "eseries-kernel-interface: error region initialized at 0x%p\n", error_info_addr);

    /* initialize metadata */
    metadata->next_id = 1;
    metadata->overflow = 0;
    metadata->conringc = 0;
    metadata->conringp = 0;

    mce_printk(KERN_INFO, "Error_info: region initialized at 0x%p\n", error_info_addr);
}

/*
 * Initialize error logging region and metadata.
 *
 *  kAddr - virtual address of error region.
 *  kLen -  available space for error region.
 */
int
init_logging_region
    (
    struct error_region* region,
    bool init_mb_error
    )
{
    struct ecc_error init_err = { 0 };

    if (!region)
    {
        printk(KERN_ERR "Invalid error region address\n");
        return -1;
    }

    error_info_addr = region;
    printk(KERN_INFO "EKI Logging region mapped at 0x%p\n", error_info_addr);

    metadata = (err_metadata*)(&region->logging[0]);

    /* Check if we should initialize the error logging region */
    if ( (strncmp(region->signature, errorRegionSignature, sizeof(region->signature)) != 0)
      || region->version != errorRegionVersion
      || metadata->error == 0
      || metadata->next_id == 0 )
    {
        printk(KERN_WARNING "Error_info: initializing region.\n");
        initialize_region();
    }

    /* This is a virtual address and may change each boot */
    metadata->error = (err_entry*) ((unsigned long)error_info_addr + sizeof(err_entry));

    /*
     * Log an entry if region was reinitialized due to uncorrectable error.
     * An explicit log is needed here because any log_error calls from the mca_logout
     * function will have been dropped due to the region not being ready yet.
     */
    if ( init_mb_error )
    {
        /* indicate that the error occurred in the persistent error region (doesn't matter where) */
        init_err.maddr = (uint64_t) persistent_region_phys_addr;
        init_err.is_uncorrectable = 1;
        log_error(0, ERROR_TYPE_ECC, &init_err);
    }

    BUILD_BUG_ON(sizeof(struct error_region) != ERROR_REGION_SIZE); /* make sure padding was adjusted if needed */
    BUILD_BUG_ON(sizeof(err_entry) != ERROR_ENTRY_SIZE); /* make sure padding was adjusted if needed */
    BUILD_BUG_ON(sizeof(err_metadata) > sizeof(err_entry));

    return 0;
}
