/*******************************************************************************

NAME            misc_ops.c

        Copyright 2016-2018 NetApp, Inc. All Rights Reserved.

*******************************************************************************/
#include <linux/types.h>
#include <linux/pci.h>
#include <linux/pci_ids.h>
#include <linux/uaccess.h>
#include <asm/pci_x86.h>
#include <asm/msr.h>

#include "eseries_kernel_interface.h"

/* MSR_PLATFORM_INFO bits */
#define PRG_TJ_OFFSET_EN        0x0000000040000000ULL

/* NETAPP PCI IDs */
#define PCI_VENDOR_ID_NETAPP 0x1275
#define PCI_DEVICE_ID_FPGA   0x7110

/* CIP interrupt control register */
#define CIP_INT_CTRL 0x14C

/* PCI configuration address defines for the Global Error module */
#define GBL_ERROR_DEV 5

/* NTB related defines and macros */
#define NTB_UPSTREAM_PORT_CONTROL  0x208
#define DISABLE_PORT_0             0x00000001
#define PLX_PORT_DATA_SIZE 4096

#define NTB_PORT_BASE_ADDR(dev) (PCI_SLOT(dev->bus->self->devfn) * PLX_PORT_DATA_SIZE)

/* NOTE: This count must match the equivalent value in RAID. */
#define TSOD_DEVICE    0x13
const int TSOD_BRIDGE[NUM_TSOD_DEVS] = { 0x180, 0x190 };

#define SMB_STAT      0x00
#define SMB_STAT_BUSY (1 << 28)

#define SMB_CTRL              0x08
#define SMB_CTRL_DTI_MASK     0xF0000000
#define SMB_CTRL_DTI_EEPROM   0xA0000000
#define SMB_CTRL_TSOD_POLL_EN 0x00000100
#define SMB_CTRL_CKOVRD       0x08000000

/**
 * \brief Get MMCONFIG address.
 */
int
eki_pci_get_mmconfig
    (
    unsigned long arg
    )
{
    int                       ret;
    struct pci_mmcfg_region*  cfg;
    struct eki_mmconfig       mmconfig_data;

    /* Lookup mmconfig for domain 0, bus 0 */
    cfg = pci_mmconfig_lookup(0, 0);
    mmconfig_data.mmconfigAddr = cfg->address;
    printk(KERN_INFO "eki_pci_get_mmconfig: address 0x%llx\n", mmconfig_data.mmconfigAddr);

    ret = copy_to_user((struct eki_mmconfig*)arg, &mmconfig_data, sizeof(struct eki_mmconfig));
    if (ret)
    {
        printk(KERN_ERR "eki_pci_get_mmconfig: can't copy data to user space, %d\n", ret);
        ret = -EFAULT;
        return ret;
    }

    return 0;
}

/**
 * \brief Get the CPU frequency.
 *
 * \param[out] arg Userspace memory to copy the CPU frequency into.
 *
 * \returns 0 if the operation was successful, otherwise negative errno.
 */
int
eki_cpufreq_get
    (
    unsigned long arg
    )
{
    uint64_t platMsr;
    uint64_t cpuFreq;

    rdmsrl(MSR_PLATFORM_INFO, platMsr);
    cpuFreq = ((platMsr >> 8) & 0xff) * 100000000;

    return (copy_to_user((uint64_t*)arg, &cpuFreq, sizeof(cpuFreq)) == 0) ? 0 : -EFAULT;
}

/**
 * \brief Get the CPU temperature relative to the TCC temperature.
 *
 * \param[out] arg Userspace memory to copy the temperature into.
 *
 * \returns 0 if the operation was successful, otherwise negative errno.
 */
int
eki_relative_cputemp_get
    (
    unsigned long arg
    )
{
    uint64_t statMsr;
    uint8_t relTemp;

    rdmsrl(MSR_IA32_PACKAGE_THERM_STATUS, statMsr);
    relTemp = (statMsr >> 16) & 0x7f;

    return (copy_to_user((uint8_t*)arg, &relTemp, sizeof(relTemp)) == 0) ? 0 : -EFAULT;
}

/**
 * \brief Get the TCC temperature.
 *
 * \param[out] arg Userspace memory to copy the temperature into.
 *
 * \returns 0 if the operation was successful, otherwise negative errno.
 */
int
eki_tcctemp_get
    (
    unsigned long arg
    )
{
    uint64_t targMsr;
    uint64_t platMsr;
    uint8_t tccTemp;

    rdmsrl(MSR_IA32_TEMPERATURE_TARGET, targMsr);
    tccTemp = (targMsr >> 16) & 0xff;

    /* If an offset from the TCC is supported, use it */
    rdmsrl(MSR_PLATFORM_INFO, platMsr);
    if (platMsr & PRG_TJ_OFFSET_EN)
        tccTemp -= (targMsr >> 24) & 0x3f;

    return (copy_to_user((uint8_t*)arg, &tccTemp, sizeof(tccTemp)) == 0) ? 0 : -EFAULT;
}

/**
 * Does 2 byte read/modify/write to a PCI device's config header and clears/sets the specified
 * bits.
 *
 * \note The clear is done first. If the same bits are set in setbits and clearbits, then the
 *       result will be to set those bits.
 */
void
pci_cfg_rmw_word
    (
    struct pci_dev* dev,
    int offset,
    uint16_t setBits,
    uint16_t clearBits
    )
{
    uint16_t initialVal = 0;

    if (!dev)
        return;

    pci_read_config_word(dev, offset, &initialVal);
    pci_write_config_word(dev, offset, (initialVal & ~clearBits) | setBits);
}

/*
 * \brief Force NTB link down prior to shutting down PCI devices
 */
static void
ntbForceLinkDown
    (
    struct pci_dev *pdev
    )
{
    void __iomem* bar = 0;
    u32 offset = 0;
    u32 value = 0;

    bar = pci_iomap(pdev, 0, 0);
    if (bar == 0)
    {
        pr_err("Unable to create virtual mapping cookie for NTB BAR0");
        return;
    }

    offset = NTB_PORT_BASE_ADDR(pdev) + NTB_UPSTREAM_PORT_CONTROL;
    value = ioread32(bar + offset);
    iowrite32(value | DISABLE_PORT_0, bar + offset);
}

/**
 * Does 4 byte read/modify/write to a PCI device's config header and clears/sets the specified
 * bits.
 *
 * \note The clear is done first. If the same bits are set in setbits and clearbits, then the
 *       result will be to set those bits.
 */
void
pci_cfg_rmw_dword
    (
    struct pci_dev* dev,
    int offset,
    uint32_t setBits,
    uint32_t clearBits
    )
{
    uint32_t initialVal = 0;

    if (!dev)
        return;

    pci_read_config_dword(dev, offset, &initialVal);
    pci_write_config_dword(dev, offset, (initialVal & ~clearBits) | setBits);
}

/**
 * \brief Shutdown PCI devices (excluding the FPGA and bridge devices)
 *
 * \returns 0 if the operation was successful, otherwise negative errno.
 */
int
eki_pci_shutdown
    (
    void
    )
{
    struct pci_dev *dev = 0;
    uint16_t mask = PCI_COMMAND_MASTER | PCI_COMMAND_MEMORY | PCI_COMMAND_IO;

    dev = pci_get_domain_bus_and_slot(0, 0, PCI_DEVFN(GBL_ERROR_DEV, 0));
    if (!dev)
    {
        pr_err("Unable to find GBL_ERROR_DEV device");
        return -ENXIO;
    }

    /* Disable NMIs & clear incoming external NMI sources */
    pci_cfg_rmw_word(dev, CIP_INT_CTRL, 0x0100, 0);
    pci_dev_put(dev);
    dev = 0;

    for_each_pci_dev(dev)
    {
        /* Force NTB link down prior to reboot. */
        if ((dev->vendor == PCI_VENDOR_ID_PLX) && (dev->device == PCI_DEVICE_ID_PLX_NTB))
            ntbForceLinkDown(dev);

        /* Don't touch the bridges or the FPGA (needed to reset the board). */
        if (!((dev->class >> 16) == PCI_BASE_CLASS_BRIDGE)
            && !(dev->vendor == PCI_VENDOR_ID_NETAPP && dev->device == PCI_DEVICE_ID_FPGA))
        {
            pr_info("Shutting down PCI device: %s", pci_name(dev));
            pci_cfg_rmw_word(dev, PCI_COMMAND, 0, mask);
        }
    }

    return 0;
}

/**
 * \brief Routine to disable TSOD.
 *
 * \param[out] arg The original TSOD values (used for restoring).
 * \returns 0 if the operation was successful, otherwise negative errno.
 *
 * \note Assumes this is called on an Intel platform.
 */
int
eki_tsod_disable
    (
    unsigned long arg
    )
{
    uint32_t clearMask = SMB_CTRL_DTI_MASK | SMB_CTRL_TSOD_POLL_EN;
    uint32_t setMask = SMB_CTRL_DTI_EEPROM | SMB_CTRL_CKOVRD;
    struct pci_dev *dev = 0;
    unsigned int i = 0;
    ESC_TSOD_STATE state;
    int status = 0;

    if (copy_from_user(&state, (uint32_t*)arg, sizeof(state)) != 0)
    {
        pr_err("Unable to copy user TSOD data.");
        return -EFAULT;
    }

    dev = pci_get_domain_bus_and_slot(0, 0xff, PCI_DEVFN(TSOD_DEVICE, 0));
    if (!dev)
    {
        pr_err("Unable to find TSOD_DEVICE device");
        return -ENXIO;
    }

    /*
     * Disable TSOD polling, ensure DTI is set for EEPROM access, and ensure clock override is
     * is "released" to allow normal operation of CMD.
     */
    for (i = 0; i < NUM_TSOD_DEVS; i++)
    {
        uint32_t smbStat = 0;

        /* If already disabled, then go to next device */
        if (state.disabled[i])
            continue;

        /* If the device is busy, then simply set the overall status */
        /* to busy and go to the next device */
        pci_read_config_dword(dev, TSOD_BRIDGE[i] + SMB_STAT, &smbStat);
        if (smbStat & SMB_STAT_BUSY)
        {
            status = -EBUSY;
            continue;
        }

        /* This causes a duplicate read, it could be optimized later if needed. */
        pci_read_config_dword(dev, TSOD_BRIDGE[i] + SMB_CTRL, &state.control[i]);
        pci_cfg_rmw_dword(dev, TSOD_BRIDGE[i] + SMB_CTRL, setMask, clearMask);
        state.disabled[i] = 1;
    }

    pci_dev_put(dev);

    if (copy_to_user((uint32_t*)arg, &state, sizeof(state)) != 0)
    {
        pr_err("Unable to copy kernel TSOD data to user.");
        return -EFAULT;
    }

    return status;
}

/**
 * \brief Routine to restore TSOD.
 *
 * \param[in] arg The TSOD values to restore.
 * \returns 0 if the operation was successful, otherwise negative errno.
 *
 * \note Assumes this is called on an Intel platform.
 */
int
eki_tsod_restore
    (
    unsigned long arg
    )
{
    struct pci_dev *dev = 0;
    unsigned int i = 0;
    ESC_TSOD_STATE state;

    if (copy_from_user(&state, (uint32_t*)arg, sizeof(state)) != 0)
    {
        pr_err("Unable to copy user TSOD data.");
        return -EFAULT;
    }

    dev = pci_get_domain_bus_and_slot(0, 0xff, PCI_DEVFN(TSOD_DEVICE, 0));
    if (!dev)
    {
        pr_err("Unable to find TSOD_DEVICE device");
        return -ENXIO;
    }

    for (i = 0; i < NUM_TSOD_DEVS; i++)
    {
        if (state.disabled[i])
            pci_write_config_dword(dev, TSOD_BRIDGE[i] + SMB_CTRL, state.control[i]);
    }

    pci_dev_put(dev);

    return 0;
}
