/*******************************************************************************

    nmihandler.c
    Copyright 2016-2019 NetApp, Inc. All Rights Reserved.

    EKI NMI handling code

*******************************************************************************/
#include <linux/nmi.h>
#include <linux/eventfd.h>
#include <linux/uaccess.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/smp.h>

#include "eki_common.h"
#include "eseries_kernel_interface.h"


/* Broadwell-DE Global Error Registers (Bus 0, Dev 5, Func 2) */
#define PCI_GNERRST     0x1C0
#define PCI_GFERRST     0x1C4
#define PCI_GCERRST     0x1A8
#define BDW_ERR_DEV_ID  0x6F2A
/* Skylake D IIO Global Error Registers (Bus 0, Dev 8, Func 0) */
/* Bits 0,1,2 indicate GF,GNF,GC */
#define IIO_GSYSST      0x204
/* Bits 0,1,2,3,4,5 indicate IIO stack source 0=C-stack, 1-4=P-Stack1-4, 5=VPP stack */
/* If all stacks indicate "PCI" error, then can we get away with just checking IIO_GSYSST ??? */
#define IIO_GF_ERRST    0x218
#define IIO_GNF_ERRST   0x224
#define IIO_GC_ERRST    0x230
#define SKL_ERR_DEV_ID  0x2014

struct nmi_eventfd
{
    int                         coreNums;               /* number of VKI Cores told by user space */
    struct eventfd_ctx**        eventfd;                /* user space notification eventfd */
};
static struct  nmi_eventfd      nmiEfds;
static int     nmi_pci_handler_registered = 0;
static struct  pci_dev          *errReg = NULL;
static bool isPciError(void);

/**
 * NMI handler routine
 *
 * This routine will issue a eventfd signal to user space when a NMI
 * is issued.
 *
 * When a NMI is issued, this routine will be called back by kernel's
 * default NMI handler: default_do_nmi().
 */
int
nmi_pci_handler(unsigned int cmd, struct pt_regs *regs)
{
    int i;

    if (!isPciError())
    {
        /* This event is not PCI related, so kernel's other registered NMI LOCAL/IO/UNKNOWN
         * handlers will be called.
         */
        return NMI_DONE;
    }

    /* any issue happened during ioctl will reset nmiEfds to NULL */
    if (nmiEfds.coreNums == 0)
    {
        printk(KERN_ERR "nmi_pci_handler: no eventfds, ignore.\n");
        /* We are not ready, and return 0, so kernel's other registered NMI LOCAL/IO/UNKNOWN
         * handlers will be called.
         */
        return NMI_DONE;
    }
    else
    {
        /* TODO EFALCON-375
         * Investigate not triggering NMI on all cores. */
        if (smp_processor_id() == 0)
        {
            printk(KERN_INFO "nmi_pci_handler: CPU#0 signal eventfds to user space\n");
            for (i = 0; i < nmiEfds.coreNums; i++)
            {
                eventfd_signal(nmiEfds.eventfd[i], 1);
            }
        }
        else
        {
            printk(KERN_INFO "nmi_pci_handler: CPU#%d NMI handled\n", smp_processor_id());
        }


        /* Return 1, so kernel's other registered NMI PCI/IO/UNKNOWN handlers
         *  will not be called.
         */
        return NMI_HANDLED;
    }
}

/**
 * nmi handler registration/un-registration routine
 *
 * enable == 1: register nmi handler
 * enable == 0: unregister nmi handler
 */
int nmi_pci_handler_registration(int enable)
{
    int ret = 0;

    if (enable)
    {
        if (cpu == BROADWELL)
        {
            /* Note: pci_get_device is not safe from NMI context */
            errReg = pci_get_device(PCI_VENDOR_ID_INTEL, BDW_ERR_DEV_ID, NULL);
            if (errReg == NULL)
            {
                printk(KERN_ERR "NMI: cannot obtain global error status pci_dev\n");
                return -ENXIO;
            }
        }
        else if (cpu == SKYLAKE)
        {
            /* Note: pci_get_device is not safe from NMI context */
            errReg = pci_get_device(PCI_VENDOR_ID_INTEL, SKL_ERR_DEV_ID, NULL);
            if (errReg == NULL)
            {
                printk(KERN_ERR "NMI: cannot obtain global error status pci_dev\n");
                return -ENXIO;
            }
        }

        printk(KERN_INFO "registering eki-nmi handler ...\n");
        /* Using NMI_FLAG_FIRST so our NMI handler will be called firstly */
        ret = register_nmi_handler(NMI_LOCAL, nmi_pci_handler, NMI_FLAG_FIRST, "nmi_pci_handler");
        if (ret)
        {
            printk(KERN_ERR "Unable to register NMI_SERR handler - nmi_pci_handler: %d\n", ret);
        }
        else
        {
            nmi_pci_handler_registered = 1;
        }
    }
    else
    {
        printk(KERN_INFO "Unregistering eki-nmi handler\n");
        unregister_nmi_handler(NMI_LOCAL, "nmi_pci_handler");
        nmi_pci_handler_registered = 0;
        nmiEfds.coreNums = 0;
        if (nmiEfds.eventfd)
        {
            kfree(nmiEfds.eventfd);
            nmiEfds.eventfd = 0;
        }
    }

    return ret;
}

/**
 * eventfds obtain routine
 *
 * ioctl wrapper function to obtain
 * eventfds from user space.
 */
int nmi_pci_get_efds(unsigned long arg)
{
    int     i;
    int     ret;
    struct  eki_eventfd_data fds;
    int*    efd;

    if (nmi_pci_handler_registered == 0)
    {
        printk(KERN_ERR "eki-NMI handler: nmi_pci_handler isn't registered");
        return -EFAULT;
    }

    ret = copy_from_user(&fds, (struct eki_eventfd_data*)arg, sizeof(struct eki_eventfd_data));
    if (ret)
    {
        printk(KERN_ERR "eki-NMI handler: can't copy data from user space, %d\n", ret);
        /* if user space recall ioctl, but with some issues,
         * free nmiEfds set from previous ioctl call, as a
         * result, we will not send out nmi signal if any
         * issue happens during ioctl.
         */
        if (nmiEfds.coreNums)
        {
            nmiEfds.coreNums = 0;
            kfree(nmiEfds.eventfd);
            nmiEfds.eventfd = 0;
        }
        return -EFAULT;
    }

    /* Free up any prior data */
    if (nmiEfds.eventfd)
    {
        nmiEfds.coreNums = 0;
        kfree(nmiEfds.eventfd);
        nmiEfds.eventfd = 0;
    }

    nmiEfds.eventfd = (struct eventfd_ctx**)kmalloc(sizeof(struct eventfd_ctx*) * fds.numVKICores, GFP_KERNEL);
    if (!nmiEfds.eventfd)
    {
        printk(KERN_ERR "eki-NMI handler: cannot allocate memory for nmi_eventfd\n");
        return -ENOMEM;
    }

    efd = (int *)kmalloc(sizeof(int) * fds.numVKICores, GFP_KERNEL);
    if (!efd)
    {
        printk(KERN_ERR "eki-NMI handler: cannot allocate memory for nmi_eventfd\n");
        kfree(nmiEfds.eventfd);
        nmiEfds.eventfd = 0;
        return -ENOMEM;
    }

    ret = copy_from_user(efd, fds.efd, sizeof(int) * fds.numVKICores);
    if (ret)
    {
        printk(KERN_ERR "eki-NMI handler: can't copy efd data from user space, %d\n", ret);
        /* if user space recall ioctl, but with some issues,
         * free nmiEfds set from previous ioctl call, as a
         * result, we will not send out nmi signal if any
         * issue happens during ioctl.
         */
        kfree(nmiEfds.eventfd);
        nmiEfds.eventfd = 0;
        kfree(efd);
        return -EFAULT;
    }

    printk(KERN_INFO "received %d eventfds from user space\n", fds.numVKICores);
    for (i = 0; i < fds.numVKICores; i++)
    {
        nmiEfds.eventfd[i] = eventfd_ctx_fdget(efd[i]);
        if (IS_ERR(nmiEfds.eventfd[i]))
        {
            ret = PTR_ERR(nmiEfds.eventfd[i]);
            printk(KERN_ERR "eki-NMI handler: Can't get eventfd on VKI core %d: %d\n",
                            i, ret);
            kfree(nmiEfds.eventfd);
            nmiEfds.eventfd = 0;
            kfree(efd);
            return -EINVAL;
        }
    }
    /* give a sign to nmi_pci_handler that eventfds are ready */
    nmiEfds.coreNums = fds.numVKICores;

    /* Free up temporary memory of the passed in EFDs */
    kfree(efd);

    return ret;
}

/**
 * pci error judgement routine
 *
 * read three global error regs
 * to see whether there is a pci error
 */
static bool isPciError()
{
    u32 fatalErrVal = 0;
    u32 uncorErrVal = 0;
    u32 corErrVal = 0;

    /* sandybridge doesn't have correctable error status register, so on Soyuz
     * we don't know whether it is a correctable pci error issue for a NMI, even
     * though both fatal and uncorrect register are zeros. As a result, we just
     * return true.
     */
    if (errReg == NULL)
    {
        printk(KERN_INFO "NMI: assumed PCI error\n");
        return true;
    }

    if (cpu == BROADWELL)
    {
        pci_read_config_dword(errReg, PCI_GFERRST, &fatalErrVal);
        pci_read_config_dword(errReg, PCI_GNERRST, &uncorErrVal);
        pci_read_config_dword(errReg, PCI_GCERRST, &corErrVal);
    }
    else if (cpu == SKYLAKE)
    {
        pci_read_config_dword(errReg, IIO_GF_ERRST, &fatalErrVal);
        pci_read_config_dword(errReg, IIO_GNF_ERRST, &uncorErrVal);
        pci_read_config_dword(errReg, IIO_GC_ERRST, &corErrVal);
    }

    if (fatalErrVal || uncorErrVal || corErrVal)
    {
        printk(KERN_INFO "NMI: detected PCI error\n");
        return true;
    }

    return false;
}

