/*******************************************************************************
 *
 *   persistent_region.c
 *   Copyright 2016-2019 NetApp, Inc. All Rights Reserved.
 *
 *   Persistent region management
 *
 *******************************************************************************/

#include <linux/string.h>
#include <linux/uaccess.h>

#include "ecc.h"
#include "logging.h"
#include "persistent_region.h"

static int scrub_persistent_region(void);
static void initialize_region(void);

void*           persistent_region_addr;
unsigned long   persistent_region_len;
unsigned long   persistent_region_phys_addr;

uint32_t  regionVersion                            = REGION_VERSION;
char      regionSignature[PR_SIGNATURE_LEN]        = REGION_SIGNATURE;
uint32_t  errorRegionVersion                       = ERROR_REGION_VERSION;
char      errorRegionSignature[PR_SIGNATURE_LEN]   = ERROR_REGION_SIGNATURE;

/*
 * Initialize persistent region and metadata.
 * kAddr - virtual address of persistent region.
 * kLen -  available space for persistent region.
 */
int
init_persistent_region
    (
    void *kAddr,
    unsigned long kLen,
    unsigned long kPhysAddr
    )
{
    bool init_mb_error;
    struct persistent_region* region;
    struct region_header* header;

    /* Do not allow logging if the region is too small */
    if (kLen < sizeof(struct persistent_region))
    {
        printk(KERN_ERR "Insufficient persistent region size %lu\n", kLen);
        return -1;
    }
    if (!kAddr)
    {
        printk(KERN_ERR "Invalid error region address\n");
        return -1;
    }

    persistent_region_addr = kAddr;
    persistent_region_phys_addr = kPhysAddr;
    persistent_region_len = kLen;
    printk(KERN_INFO "EKI Persistent region mapped at 0x%lx/0x%p\n",
                      persistent_region_phys_addr, persistent_region_addr);

    init_mb_error = (bool)scrub_persistent_region();

    /* assign to a temporary pointer so that we can initialize the region first */
    region = (struct persistent_region*)persistent_region_addr;
    header = &region->header;

    /* Check if we should initialize the persistent region */
    if ( (strncmp(header->signature, regionSignature, sizeof(header->signature)) != 0)
      || header->version != regionVersion
      || init_mb_error )
    {
        printk(KERN_WARNING "eseries-kernel-interface: initializing persistent region.\n");
        initialize_region();
    }

    init_logging_region(&region->err_logging, init_mb_error);

    return 0;
}

static void
initialize_region
    (
    )
{
    struct region_header* header;

    if (!persistent_region_addr)
    {
        printk(KERN_ERR "eseries-kernel-interface: No pointer provided for initialize_region()\n");
        return;
    }

    memset(persistent_region_addr, 0, persistent_region_len);

    /* initialize metadata */
    header = (struct region_header*)persistent_region_addr;
    memcpy(header->signature, regionSignature, sizeof(header->signature));
    header->version = regionVersion;

    printk(KERN_INFO "eseries-kernel-interface: region initialized at 0x%p\n", persistent_region_addr);
}


static int
scrub_persistent_region
    (
    )
{
    bool patrol_status;
    bool mb_status;
    volatile uint64_t* p;
    uint32_t errStatus;
    int retval = 0;

    /* Disable ECC and check for errors in region */
    patrol_status = enable_patrol_scrub(0);
    mb_status = mce_mb_enabled();
    mce_enable_mb(0);

    for( p = (uint64_t*) persistent_region_addr;
          ((unsigned long) p) < ((unsigned long)persistent_region_addr + persistent_region_len);
          p++ )
    {
        (void) *p;
    }

    /* check for errors */
    errStatus = check_for_errors(false);

    printk(KERN_INFO "Persistent region returned 0x%u after scan\n", errStatus);

    /* If uncorrectable error happened, then scrub this return and return 1 to indicate */
    /* that the region was scrubbed */
    if (errStatus & (ECC_UNCOR | MCE_UNCOR | MCE_UNKNOWN))
    {
        retval = 1;
        memset(persistent_region_addr, 0, persistent_region_len);
        errStatus = check_for_errors(false);
    }

    /* restore patrol */
    mce_enable_mb(mb_status);
    enable_patrol_scrub(patrol_status);

    return retval;
}

/*
 * This function is 'called' using the ioctl interface(EKI_IOCTL_PER_REG_IMT). Persistent region
 * IMT tests access necessary data using this function. arg is pointer to the structure
 * 'persistent_reg_imtdata'. And the user is responsible to clear the structure member
 * 'signature' before using it.
 */
int eki_per_reg_imt(unsigned long arg)
{
    int ret;
    struct persistent_reg_imtdata imtData;
    struct persistent_region* region;
    struct region_header* header;
    struct error_region* err_region;

    ret = copy_from_user(&imtData, (struct persistent_reg_imtdata*)arg, sizeof(struct persistent_reg_imtdata));
    if (ret)
    {
        printk(KERN_ERR "eki_per_reg_imt: can't copy data from user space, %d\n", ret);
        ret = -EFAULT;
        return ret;
    }

    switch (imtData.op)
    {
        /* read the version/signature from header and error logging persistent memory */
        case READ_REGION:
            region = (struct persistent_region*) persistent_region_addr;
            switch(imtData.var)
            {
                case REG_VER:
                  header = (struct region_header*)(&region->header);
                  imtData.data.version = header->version;
                  break;
                case REG_SIG:
                  header = (struct region_header*)(&region->header);
                  memcpy(imtData.data.signature, header->signature, sizeof(header->signature));
                  break;
                case ERR_REG_VER:
                  err_region = (struct error_region*)(&region->err_logging);
                  imtData.data.version = err_region->version;
                  break;
                case ERR_REG_SIG:
                  err_region = (struct error_region*)(&region->err_logging);
                  memcpy(imtData.data.signature, err_region->signature, sizeof(err_region->signature));
                  break;
                default:
                  return EINVAL;
            }
          break;
          /* read the in-memory version/signature global variables */
        case READ_GLOBAL:
          switch(imtData.var)
            {
                case REG_VER:
                  imtData.data.version = regionVersion;
                  break;
                case REG_SIG:
                  memcpy(imtData.data.signature, regionSignature, sizeof(regionSignature));
                  break;
                case ERR_REG_VER:
                  imtData.data.version = errorRegionVersion;
                  break;
                case ERR_REG_SIG:
                  memcpy(imtData.data.signature, errorRegionSignature, sizeof(errorRegionSignature));
                  break;
                default:
                  return EINVAL;
            }
          break;
          /* write to the in-memory version/signature global variables */
        case WRITE_GLOBAL:
          switch(imtData.var)
            {
                case REG_VER:
                  regionVersion = imtData.data.version;
                  break;
                case REG_SIG:
                  memset(regionSignature, '\0', PR_SIGNATURE_LEN);
                  memcpy(regionSignature, imtData.data.signature, sizeof(regionSignature));
                  break;
                case ERR_REG_VER:
                  errorRegionVersion = imtData.data.version;
                  break;
                case ERR_REG_SIG:
                  memset(errorRegionSignature, '\0', PR_SIGNATURE_LEN);
                  memcpy(errorRegionSignature, imtData.data.signature, sizeof(errorRegionSignature));
                  break;
                default:
                  return EINVAL;
            }
          return 0;
          break;
        /* call the persistent region initialization function */
        case CALL_INIT:
           return init_persistent_region(persistent_region_addr, persistent_region_len,
                                         persistent_region_phys_addr);
          break;
        default:
            printk("Invalid operation %d in the argument\n", imtData.op);
            return EINVAL;
    }

    ret = copy_to_user((struct presistent_reg_imtdata*)arg, &imtData, sizeof(struct persistent_reg_imtdata));
    if (ret)
    {
        printk(KERN_ERR "eki_per_reg_imt: can't copy data to user space, %d\n", ret);
        ret = -EFAULT;
        return ret;
    }

    return 0;
}
