/*******************************************************************************

NAME            rmem-mgr.c

        Copyright 2015-2018 NetApp, Inc. All Rights Reserved.

DESCRIPTION:
 This is the source file for the Reserved Memory Manager loadable kernel module.

*******************************************************************************/

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/uaccess.h>
#include <linux/string.h>
#include "rmem.h"
#include "rmem-mgr.h"

/* Newer kernels don't define _PAGE_CACHE_WB but instead have a translator */
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,9,0)
    #define _PAGE_CACHE_WB  cachemode2protval(_PAGE_CACHE_MODE_WB)
#endif

#undef DEBUG_MSG

#define RMEM_VERSION "1.0"
#define RMEM_MAX_MINOR RMEM_MAX_NUM_DEVICES

#define RMEM_APP_DEV_CLEANUP  1
#define RMEM_FULL_DEV_CLEANUP 2

/* Total reserved memory map parameter */
static char *memmap = NULL;
module_param_named(reserved, memmap, charp, S_IRUGO);

/* Kernel reserved persisted memory parameter */
static char *kernelMemMap = NULL;
module_param_named(klog, kernelMemMap, charp, S_IRUGO);

static dev_t rmem_major;
static struct class* rmem_class;
static int ndevices;
static void* kmemAddr;
static unsigned long kmemPhysAddr;
static unsigned long kmemLength;


struct rmem_dev
{
    unsigned long base_addr;
    unsigned long length;
    unsigned long write_max;
    unsigned long read_max;
    unsigned long read_offset;
    unsigned long write_offset;
    int is_core_dumped;
    int use_offset;
    int allow_file_operations;
    char name[RMEM_DEVNAME_LEN];
    struct cdev cdev;
};

static struct rmem_dev rdevlist[RMEM_MAX_MINOR];
static eki_clear_memory_t clear_memory_func = NULL;

static void rmem_cleanup(int cleanup);
static int rmem_create_devices(struct rmem_devices* deviceInfo);

/**
 * rmem returns address and size of kernel log space
 *
 */
void
rmem_register_clear_memory
    (
    eki_clear_memory_t eki_clear_memory
    )
{
    clear_memory_func = eki_clear_memory;
}
EXPORT_SYMBOL(rmem_register_clear_memory);

/**
 * rmem returns address and size of kernel log space
 *
 */
void*
rmem_kernel_log_memory
    (
    unsigned long* length,
    unsigned long* physAddr
    )
{
    if (length != 0)
        *length = kmemLength;
    if (physAddr != 0)
        *physAddr = kmemPhysAddr;
    return kmemAddr;
}
EXPORT_SYMBOL(rmem_kernel_log_memory);

/**
 * rmem populates the address and size of eos and rpa memory
 *
 */
void
rmem_eos_rpa_memory
    (
    unsigned long* eosAddr,
    unsigned long* eosLength,
    unsigned long* rpaAddr
    )
{
    int minor;

    if (eosAddr != 0)
        *eosAddr = rdevlist[0].base_addr;
    if (eosLength != 0)
        *eosLength = rdevlist[0].length;

    /* Skip the default rmem device which is the first device in the rdevlist */
    for (minor = 1; minor < ndevices; minor++)
    {
        if (strncmp(rdevlist[minor].name, RMEM_RPA_REGION_START,
                sizeof(rdevlist[minor].name)) == 0)
        {
            if (rpaAddr != 0)
                *rpaAddr = rdevlist[minor].base_addr;
        }
    }
}
EXPORT_SYMBOL(rmem_eos_rpa_memory);

/**
 * rmem device open routine.
 *
 * This function figures out the instances of this driver and stores
 * the private data structures into the file pointer for this device
 * to be used by the other fileOps.
 *
 */
static int
rmem_open(struct inode* inode, struct file* filp)
{
    unsigned int major = imajor(inode);
    unsigned int minor = iminor(inode);
    struct rmem_dev* dev = &rdevlist[minor];

#ifdef DEBUG_MSG
    printk(KERN_DEBUG "/dev/%s is open\n", dev->name);
#endif

    if (major != rmem_major || minor < 0 || minor > ndevices)
    {
        printk(KERN_ERR "No device found with major: %d and minor: %d\n", major, minor);
        return -ENODEV;
    }

    if (inode->i_cdev != &dev->cdev)
    {
        printk(KERN_ERR "open: internal error\n");
        return -ENODEV;
    }

    if(dev->use_offset)
    {
        dev->write_offset = (dev->write_max == 0) ? 0 : dev->write_max - 1;
        dev->read_offset = (dev->read_max == 0) ? 0 : dev->read_max - 1;

        if(filp->f_mode & FMODE_WRITE)
        {
            filp->f_pos = dev->write_offset;
        }
        else if(filp->f_mode & FMODE_READ)
        {
            filp->f_pos = dev->read_offset;
        }
    }

    filp->private_data = dev;
    return 0;
}

/**
 * This function will copy data from a buffer(eos device) to user space
 */
static ssize_t
rmem_read(struct file* filep, char __user* buffer, size_t length, loff_t* offset)
{
    char* remap_ptr = 0;
    ssize_t bytes_read     = 0;
    loff_t localOffset     = 0;
    loff_t devOffset       = 0;

    struct rmem_dev* dev = (struct rmem_dev*)filep->private_data;

    if (length == 0)
        return 0;

    if (dev->allow_file_operations)
    {
        /*
         * If use offsets is set we need to use the device read offset for reads.
         */
        if(dev->use_offset)
        {
            devOffset = dev->read_offset;

            /* If we try to read passed what we have written truncate the read */
            if ((devOffset + length) > dev->write_offset)
            {
                length = dev->write_offset - dev->read_offset;
            }

            if (length == 0)
            {
                return -EINVAL;
            }
        }
        else
        {
            devOffset = *offset;
        }

        if (dev->length < (devOffset + length))
        {
            printk(KERN_ERR "rmem_read %s: reading past the end of device, endAddr:0x%lx "
                   "requested endAddr: 0x%llx\n",
                   dev->name, dev->base_addr + dev->length, dev->base_addr + devOffset + length);
            return -EINVAL;
        }

        remap_ptr = (char *)ioremap_cache(dev->base_addr + devOffset, length);

        if (!remap_ptr)
        {
            printk(KERN_ERR "rmem_read %s: ioremap_cache failed, endAddr:0x%lx "
                   "requested endAddr: 0x%llx\n",
                   dev->name, dev->base_addr + dev->length, dev->base_addr + devOffset + length);
            return -ENXIO;
        }

        /*
         * The localOffset is 0 as we are starting at the beginning of the remapped area
         * and the read will return the number bytes read and the offset after writing.
         */
        bytes_read = simple_read_from_buffer(buffer, length, &localOffset, remap_ptr,
                                             length);
        if (bytes_read < 0)
        {
            printk(KERN_ERR "rmem_read %s: simple_read_from_buffer failed, bytes_read: %lx\n",
                   dev->name, bytes_read);
            iounmap(remap_ptr);
            return bytes_read;
        }

        devOffset += localOffset;
        filep->f_pos = devOffset;

        if (devOffset > dev->read_max)
        {
            dev->read_max = devOffset;
            if (dev->use_offset)
            {
                dev->read_offset = dev->read_max - 1;
            }
        }

        *offset = devOffset;

        iounmap(remap_ptr);
    }

    return bytes_read;
}

/**
 * This function will copy data from user space to a buffer(eos device)
 */
static ssize_t
rmem_write (struct file* filep, const char __user* buffer, size_t length, loff_t* offset)
{
    char* remap_ptr = 0;
    ssize_t bytes_written  = 0;
    loff_t localOffset     = 0;
    loff_t devOffset       = 0;

    struct rmem_dev* dev = (struct rmem_dev*)filep->private_data;

    if (length == 0)
        return 0;

    if (dev->allow_file_operations)
    {
        /* If use offsets is set we need to use the device write offset for writes. */
        if(dev->use_offset)
        {
            devOffset = dev->write_offset;

            /*
             * If this write will hit the end of the buffer make sure we set the length
             * to stop at the end of the buffer.
             */
            if ((devOffset + length) > dev->length)
            {
                length = dev->length - devOffset;
            }

            if (length == 0)
            {
                return -ENOSPC;
            }
        }
        else
        {
            devOffset = *offset;
        }

        if ((devOffset + length) > dev->length)
        {
            printk(KERN_ERR "rmem_write %s: No space left on device, endAddr:0x%lx "
                   "requested endAddr: 0x%llx devOffset = 0x%llx length = 0x%lx\n",
                   dev->name, dev->base_addr + dev->length, dev->base_addr + devOffset + length,
                   devOffset, length);
            return -ENOSPC;
        }

        remap_ptr = (char *)ioremap_cache(dev->base_addr + devOffset, length);

        if (!remap_ptr)
        {
            printk(KERN_ERR "rmem_write %s: ioremap_cache failed, endAddr:0x%lx "
                   "requested endAddr: 0x%llx devOffset = 0x%llx length = 0x%lx\n",
                   dev->name, dev->base_addr + dev->length, dev->base_addr + length + devOffset,
                   devOffset, length);
            return -ENXIO;
        }

        /*
         * The localOffset is 0 as we are starting at the beginning of the remapped area
         * and the write will return the offset after writing.
         */
        bytes_written = simple_write_to_buffer(remap_ptr, length, &localOffset,
                                               buffer, length);
        if (bytes_written < 0)
        {
            printk(KERN_ERR "rmem_write %s: simple_write_to_buffer failed, bytes_written: 0x%lx\n",
                   dev->name, bytes_written);
            iounmap(remap_ptr);
            return bytes_written;
        }

        devOffset += localOffset;

        /* Save the maximum offset to determine the size of the file */
        if (devOffset > dev->write_max)
        {
            dev->write_max = devOffset;
            if (dev->use_offset)
            {
                dev->write_offset = dev->write_max - 1;
            }
        }

        filep->f_pos = devOffset;

        *offset = devOffset;

        iounmap(remap_ptr);
    }

    return bytes_written;
}

/**
 * This function will lseek within an eos device)
 */
loff_t rmem_llseek (struct file* filep, loff_t offset, int whence)
{
    loff_t seek_offset = 0;
    struct rmem_dev* dev = (struct rmem_dev*)filep->private_data;

    if (dev->allow_file_operations)
    {
        seek_offset = fixed_size_llseek(filep, offset, whence, dev->length);

        /*
         * Need to update write and read offset depending on how the file was
         * opened.
         */
        if (dev->use_offset)
        {
            /* Need to check how we open this file. */
            if (filep->f_mode & FMODE_WRITE)
            {
                dev->write_offset = seek_offset;
                dev->write_max = dev->write_offset + 1;
            }

            if (filep->f_mode & FMODE_READ)
            {
                /* Read Only case */
                if ( seek_offset > dev->write_offset )
                {
                    return -ENOSPC;
                }

                dev->read_offset = seek_offset;
                dev->read_max = dev->read_offset + 1;
            }
        }

        return seek_offset;
    }

    return -EINVAL;
}

/**
 * rmem device close routine.
 *
 * No special operations are required for this close routine.
 */
static int
rmem_release(struct inode* inode, struct file* filp)
{
#ifdef DEBUG_MSG
    unsigned int minor = iminor(inode);
    struct rmem_dev* dev = &rdevlist[minor];

    printk(KERN_DEBUG "/dev/%s is released\n", dev->name);
#endif
    return 0;
}

static const struct vm_operations_struct rmem_physical_vm_ops =
{
#ifdef CONFIG_HAVE_IOREMAP_PROT
    .access = generic_access_phys
#endif
};

/**
 * rmem device mmap routine.
 *
 * This function remaps the assigned kernel memory of this device to userspace.
 */
static int
rmem_mmap(struct file* filep, struct vm_area_struct* vma)
{
    struct rmem_dev* dev = (struct rmem_dev *) filep->private_data;
    size_t size;
    int result;

#ifdef DEBUG_MSG
    printk(KERN_DEBUG "In rmem_mmap...\n");
#endif

    size = vma->vm_end - vma->vm_start;
    if (size != dev->length)
    {
        printk(KERN_ERR "Memory region size is wrong\n");
        return -EINVAL;
    }

    vma->vm_ops = &rmem_physical_vm_ops;

    vma->vm_page_prot =
            __pgprot((pgprot_val(vma->vm_page_prot) & ~_PAGE_CACHE_MASK) | _PAGE_CACHE_WB);

    vma->vm_pgoff = (dev->base_addr >> PAGE_SHIFT) + vma->vm_pgoff;

    result = remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff,
                             size, vma->vm_page_prot);
    if (result)
    {
        printk(KERN_ERR "Failed to remap kernel memory to userspace.\n");
        return -EAGAIN;
    }

    /*
     * Check if this device's memory should be included in the core dump.
     */
    if (dev->is_core_dumped)
    {
        vma->vm_flags &= ~VM_IO;
        vma->vm_flags &= ~VM_DONTDUMP;
        printk(KERN_DEBUG "Set vm_flags for core dump, device:%s\n", dev->name);
    }

#ifdef DEBUG_MSG
    printk(KERN_DEBUG "rmem_mmap succeeded...\n");
#endif
    return 0;
}

/**
 * rmem device ioctl routine.
 *
 * This function supports only the following ioctl commands:
 *
 * - RMEM_GET_INFO: Gets base address and length of the memory region
 * - RMEM_CREATE_DEVS: Create the given devices
 * - RMEM_SET_CORE_ADDRS: Set the starting address of mmap regions that should be in
 *                        the core dump.
 * - RMEM_CLEAR_LENGTH: Clear the variable tracking the length of the data written.
 * - RMEM_GET_LENGTH: Get the length of the data written to the device
 *
 * The following ioctl functions are used for devices that use offsets.  Currently
 * the only device that uses offsets is the rds device.  RDS is unique in that it
 * needs to use this device as both a buffer and a file.  Both of these operations
 * need to access read/write offsets in order to track what offset to use when
 * accessing this device.  File operations that go through file functions like
 * rmem_open, rmem_write, and rmem_llseek need to properly update the read and
 * write offsets.  This allows requests that use this device as a buffer to access
 * the device from the correct offset with the correct data.
 *
 * - RMEM_SET_USEOFFSET: Tell device to use offsets.
 * - RMEM_GET_FREECNT: Get the amount of free space from the device.
 * - RMEM_CLEAR_BUFFER: Clear data in buffer and reset offsets.
 * - RMEM_SET_WRTOFFSET: Set device write offset.
 * - RMEM_SET_RDOFFSET: Set device read offset.
 * - RMEM_GET_WRTOFFSET: Get device write offset.
 * - RMEM_GET_RDOFFSET: Get device read offset.
 *
 * For all other commands, it will return error.
 */
long
rmem_ioctl(struct file *filep, unsigned int cmd, unsigned long arg)
{
    struct rmem_dev* dev = (struct rmem_dev *) filep->private_data;
    struct map_info info;
    struct rmem_devices devices;
    int offset;
    int useOffsets;
    char* remap_ptr = 0;

    switch(cmd)
    {
    case RMEM_GET_INFO:

        /* This command can only be used on the RMEM default device */
        if (strnstr(dev->name, RMEM_NAME, sizeof(dev->name)) == 0)
        {
            printk(KERN_ERR "rmem_ioctl(RMEM_GET_INFO): %s: invalid device\n", dev->name);
            return -EINVAL;
        }

        info.length = dev->length;
        info.addr = dev->base_addr;
        printk(KERN_INFO "rmem_ioctl(RMEM_GET_INFO): %s: base_addr: 0x%lx length: 0x%lx\n",
               dev->name, dev->base_addr, dev->length);
        return copy_to_user((void*)arg, &info, sizeof(info));

    case RMEM_CREATE_DEVS:

        /* This command can only be used on the RMEM default device */
        if (strnstr(dev->name, RMEM_NAME, sizeof(dev->name)) == 0)
        {
            printk(KERN_ERR "rmem_ioctl(RMEM_CREATE_DEVS): %s: invalid device\n", dev->name);
            return -EINVAL;
        }

        printk(KERN_INFO "rmem_ioctl(RMEM_CREATE_DEVS)\n");
        if(copy_from_user(&devices, (void*)arg, sizeof(devices)))
        {
            printk(KERN_ERR "rmem_ioctl: copy_from_user failed\n");
            return -EFAULT;
        }
        printk(KERN_INFO "rmem_ioctl(RMEM_CREATE_DEVS) num_devices:%d\n", devices.num_devices);
        return rmem_create_devices(&devices);

    case RMEM_CLEAR_LENGTH:
        dev->write_max = 0;
        return 0;

    case RMEM_GET_LENGTH:
        return copy_to_user((void*)arg, &dev->write_max, sizeof(unsigned long));
    case RMEM_SET_USEOFFSET:
        if(copy_from_user(&useOffsets, (void*)arg, sizeof(int)))
        {
            printk(KERN_ERR "rmem_ioctl: copy_from_user failed\n");
            return -EFAULT;
        }

        if (useOffsets)
        {
            printk(KERN_INFO "rmem_ioctl(RMEM_SET_USEOFFSET) setting offsets %s\n", dev->name);
            dev->use_offset = 1;
        }
        else
            dev->use_offset = 0;

        return 0;

    case RMEM_GET_FREECNT:
        if (dev->use_offset)
        {
            int length = dev->length - dev->write_offset;

            return copy_to_user((void*)arg, &length, sizeof(int));
        }
        else
        {
            printk(KERN_ERR "rmem_ioctl(RMEM_GET_FREECNT): %s: device doesn't use offsets\n", dev->name);
            return -EINVAL;
        }

    case RMEM_CLEAR_BUFFER:
        if (dev->use_offset)
        {
            dev->write_max = 0;
            dev->write_offset = 0;
            dev->read_max = 0;
            dev->read_offset = 0;

            remap_ptr = ioremap_cache(dev->base_addr, dev->length);

            if(!remap_ptr)
            {
                printk(KERN_ERR "rmem_ioctl clear buffer %s: ioremap_cache failed, requested addr:0x%lx "
                       " length: %lu",
                       dev->name, dev->base_addr, dev->length);
                return -ENXIO;
            }

            memset(remap_ptr , 0, dev->length);
            iounmap(remap_ptr);
            return 0;
        }
        else
        {
            printk(KERN_ERR "rmem_ioctl(RMEM_CLEAR_BUFFER): %s: device doesn't use offsets\n", dev->name);
            return -EINVAL;
        }

    case RMEM_SET_WRTOFFSET:
        if (dev->use_offset)
        {
            if(copy_from_user(&offset, (void*)arg, sizeof(int)))
            {
                printk(KERN_ERR "rmem_ioctl: copy_from_user failed\n");
                return -EFAULT;
            }

            dev->write_offset = offset;
            dev->write_max = dev->write_offset + 1;
            return 0;
        }
        else
        {
            printk(KERN_ERR "rmem_ioctl(RMEM_SET_WRTOFFSET): %s: device doesn't use offsets\n", dev->name);
            return -EINVAL;
        }
    case RMEM_SET_RDOFFSET:
        if (dev->use_offset)
        {
            if(copy_from_user(&offset, (void*)arg, sizeof(int)))
            {
                printk(KERN_ERR "rmem_ioctl: copy_from_user failed\n");
                return -EFAULT;
            }

            if ((offset > dev->write_offset) && dev->write_offset)
            {
                printk(KERN_ERR "rmem_ioctl(RMEM_SET_RDOFFSET) %s writeOffset:%lu larger than argument:%d\n",
                       dev->name, dev->write_offset, offset);

                return -EINVAL;
            }

            dev->read_offset = offset;
            dev->read_max = dev->read_offset + 1;
            return 0;
        }
        else
        {
            printk(KERN_ERR "rmem_ioctl(RMEM_SET_RDOFFSET): %s: device doesn't use offsets\n", dev->name);
            return -EINVAL;
        }
    case RMEM_GET_WRTOFFSET:
        if (dev->use_offset)
        {
            offset = dev->write_offset;
            return copy_to_user((void*)arg, &offset, sizeof(int));
        }
        else
        {
            printk(KERN_ERR "rmem_ioctl(RMEM_GET_WRTOFFSET): %s: device doesn't use offsets\n", dev->name);
            return -EINVAL;
        }
    case RMEM_GET_RDOFFSET:
        if (dev->use_offset)
        {
            offset = dev->read_offset;
            return copy_to_user((void*)arg, &offset, sizeof(int));
        }
        else
        {
            printk(KERN_ERR "rmem_ioctl(RMEM_GET_RDOFFSET): %s: device doesn't use offsets\n", dev->name);
            return -EINVAL;
        }

    default:
        return -ENOTTY;
    }
}

/**
 * Standard file operations structure to hold pointers of functions
 * to handle device specific operation.
 */
static struct file_operations rmem_fops =
{
    .owner          = THIS_MODULE,
    .open           = rmem_open,
    .read           = rmem_read,
    .write          = rmem_write,
    .llseek         = rmem_llseek,
    .release        = rmem_release,
    .unlocked_ioctl = rmem_ioctl,
    .mmap           = rmem_mmap
};

/**
 * Creates the device in sysfs and notify udev daemon to create device
 * files in /dev
 */
static int
rmem_register(int minor)
{
    struct device* device;
    dev_t devno;

    devno = MKDEV(rmem_major, minor);

    cdev_init(&rdevlist[minor].cdev, &rmem_fops);
    rdevlist[minor].cdev.owner = THIS_MODULE;

    if (cdev_add(&rdevlist[minor].cdev, devno, 1) == -1)
    {
        printk(KERN_ERR "Failed to add %s\n", rdevlist[minor].name);
        return -1;
    }

    device = device_create(rmem_class, NULL, devno, NULL, rdevlist[minor].name);
    if (IS_ERR(device))
    {
        printk(KERN_ERR "Failed to create %s\n", rdevlist[minor].name);
        cdev_del(&rdevlist[minor].cdev);
        return PTR_ERR(device);
    }

    return 0;
}

static int
rmem_clear_memory
    (
    unsigned long addr,
    unsigned long length
    )
{
    char* remap_ptr = 0;

    remap_ptr = ioremap_cache(addr, length);
    if(!remap_ptr)
    {
        printk(KERN_ERR "rmem_create_devices: ioremap_cache failed,"
               " requested addr:0x%lx length: %lu\n", addr, length);
        return -ENXIO;
    }

    if (clear_memory_func)
    {
        clear_memory_func((void *)remap_ptr, length, false);
    }
    else
    {
        printk(KERN_ERR "rmem_create_devices: clear_memory_func not registered\n");
        iounmap(remap_ptr);
        return -ENXIO;
    }

    iounmap(remap_ptr);
    return 0;
}

static int
rmem_create_devices(struct rmem_devices* devices)
{
    static int minor = 1; /* rmem default device is minor 0, thus the new devices start at 1 */
    int index;
    int ret;

    if (minor > RMEM_MAX_NUM_DEVICES)
    {
        printk(KERN_ERR "rmem_create_devices numDevices:%d exceeds max devices %d\n",
               devices->num_devices, RMEM_MAX_NUM_DEVICES);
        return -1;
    }

    /* Note: The default rmem device is the first device in the rdevlist using minor 0 */
    for (index = 0; index < devices->num_devices; minor++, index++, ndevices++)
    {
        strncpy(rdevlist[minor].name, devices->devices[index].name,
                sizeof(rdevlist[minor].name));

        rdevlist[minor].length = devices->devices[index].length;
        rdevlist[minor].base_addr = devices->devices[index].addr;
        rdevlist[minor].write_max = 0;
        rdevlist[minor].is_core_dumped = devices->devices[index].is_core_dumped;
        rdevlist[minor].allow_file_operations = devices->devices[index].allow_file_operations;
        rdevlist[minor].use_offset = 0;
        rdevlist[minor].read_max = 0;
        rdevlist[minor].write_offset = 0;
        rdevlist[minor].read_offset = 0;

        if ((ret = rmem_register(minor)))
        {
            printk(KERN_ERR "rmem_create_devices: registration failed\n");
            rmem_cleanup(RMEM_APP_DEV_CLEANUP);
            return ret;
        }

        if (devices->devices[index].clear_on_init)
        {
            ret = rmem_clear_memory(devices->devices[index].addr,
                                    devices->devices[index].length);
            if (ret)
            {
                printk(KERN_ERR "rmem_create_devices: clear failed\n");
                rmem_cleanup(RMEM_APP_DEV_CLEANUP);
                return ret;
            }
        }

        printk(KERN_INFO "rmem_create_devices %s: base_addr 0x%016lx length 0x%016lx minor:%d\n",
               rdevlist[minor].name, rdevlist[minor].base_addr, rdevlist[minor].length, minor);
    }

    return 0;
}

/**
 * Module initialization function. It parses the module parameter to figure out the
 * number of devices to be created and each devices base address and length. It then
 * creates the devices and registers it with sysfs for user space application access.
 */
static int __init
rmem_init(void)
{
    unsigned long baseAddr;
    unsigned long length;
    dev_t devno;
    char* str;
    int ret;

    ndevices = 0;

    printk(KERN_INFO "rmem_init: Initializing reserved memory manager module.\n");

    if (memmap == 0)
    {
        printk(KERN_ERR "rmem_init: No argument is passed to rmem-mgr module for reserved memory size/addr.\n");
        return -EINVAL;
    }

    str = memmap;

    /* Get reserved memory region size */
    length = memparse(str, &str);
    if (*str != '$')
    {
        printk(KERN_ERR "rmem_init: Invalid string %s\n", str);
        return -EINVAL;
    }

    baseAddr = memparse(str + 1, &str);

    printk(KERN_INFO "rmem_init: Given base_addr 0x%016lx length 0x%016lx\n", baseAddr, length);

    if (kernelMemMap == 0)
    {
        printk(KERN_ERR "rmem_init: No argument is passed to rmem-mgr module for kernel log size.\n");
        return -EINVAL;
    }

    str = kernelMemMap;

    /* Get kernel memory region size */
    kmemLength = memparse(str, &str);
    printk(KERN_INFO "rmem_init: Given kernel log length of 0x%016lx\n", kmemLength);

    kmemPhysAddr = baseAddr;
    kmemAddr = (void *)ioremap_cache(baseAddr, kmemLength);
    if (!kmemAddr)
    {
        printk(KERN_ERR "rmem_init kmemAddr: ioremap_cache failed, Addr:0x%016lx "
               "requested endAddr: 0x%016lx\n", baseAddr, kmemLength);
        return -ENXIO;
    }

    /* Reduce the memory size by the size of the kernel memory size */
    baseAddr += kmemLength;
    length   -= kmemLength;

    /* Dynamically allocate major and minor device numbers. */
    /* # cat /proc/devices */
    if ((ret = alloc_chrdev_region(&devno, 0, RMEM_MAX_MINOR, RMEM_NAME)))
    {
        printk(KERN_ERR "rmem_init: Unable to get major number for rmem devices\n");
        return ret;
    }

    rmem_major = MAJOR(devno);

    /* # ls /sys/class */
    rmem_class = class_create(THIS_MODULE, RMEM_NAME);
    if (IS_ERR(rmem_class))
    {
        printk(KERN_ERR "rmem_init: class registration failed\n");
        unregister_chrdev_region(rmem_major, RMEM_MAX_MINOR);
        return PTR_ERR(rmem_class);
    }

    /* Create the default /dev/rmem device minor:0 */
    strncpy(rdevlist[0].name, RMEM_NAME, sizeof(rdevlist[0].name));
    rdevlist[0].length = length;
    rdevlist[0].base_addr = baseAddr;

    if ((ret = rmem_register(0)))
    {
        printk(KERN_ERR "rmem_init: registration failed\n");
        class_destroy(rmem_class);
        unregister_chrdev_region(MKDEV(rmem_major, 0), RMEM_MAX_MINOR);
        return ret;
    }

    ndevices = 1;

    return 0;
}

/**
 * Module cleanup routine. It will be called in case of error during initialization
 * and by module exit function. It frees up the system resources used.
 */
static void
rmem_cleanup(int cleanup)
{
    int minor = 0;

    printk(KERN_DEBUG "rmem_cleanup...\n");

    // Skip the rmem device only cleanup the application created devices
    if (cleanup == RMEM_APP_DEV_CLEANUP)
        minor = 1;

    for (; minor < ndevices; ++minor)
    {
        device_destroy(rmem_class, MKDEV(rmem_major, minor));
        cdev_del(&rdevlist[minor].cdev);
    }

    /* Resetting ndevices back to one to account for the default rmem device */
    if (cleanup == RMEM_APP_DEV_CLEANUP)
        ndevices = 1;

    if (cleanup == RMEM_FULL_DEV_CLEANUP)
    {
        class_destroy(rmem_class);
        unregister_chrdev_region(MKDEV(rmem_major, 0), RMEM_MAX_MINOR);
        ndevices = 0;
    }
}

/**
 * Module exit function
 */
static void __exit
rmem_exit(void)
{
    printk(KERN_INFO "Unloading reserved memory manager module.\n");
    rmem_cleanup(RMEM_FULL_DEV_CLEANUP);
}

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("NetApp Reserved Memory Manager" RMEM_VERSION);
MODULE_AUTHOR("NetApp, Inc");

module_init(rmem_init);
module_exit(rmem_exit);
